#!/usr/bin/python3

import os
import sys
import string
import random
import time
import socket
import re

try:
    import pwd
except ImportError:
    import getpass
    pwd = None

using_python_3 = True if sys.version_info.major == 3 else False


def get_current_user():
    if pwd:
        return pwd.getpwuid(os.geteuid()).pw_name
    else:
        return getpass.getuser()


def user_input():
    if using_python_3:
        return eval(input())
    else:
        # using python version 2
        return eval(input())


class random_id_gen:
    """
    Emulated generator for creating a random chars id of specific length

    :parameters:
        length : int
            the desired length of the generated id

            default: 8

    :return:
        a random id with each next() request.
    """

    def __init__(self, length=8):
        self.id_chars = string.ascii_lowercase + string.digits
        self.length = length

    def __next__(self):
        return ''.join(random.choice(self.id_chars) for _ in range(
            self.length))

    # __next__ = next
    next = __next__


# try to get number from input, return None in case of fail
def get_number(input):
    try:
        return int(input)
    except Exception:
        try:
            return int(input)
        except Exception:
            return None


def list_intersect(l1, l2):
    return list([x for x in l1 if x in l2])

# actually first list minus second


def list_difference(l1, l2):
    return list([x for x in l1 if x not in l2])

# symmetric diff


def list_xor(l1, l2):
    return list(set(l1) ^ set(l2))


def is_sub_list(l1, l2):
    return set(l1) <= set(l2)

# splits a timestamp in seconds to sec/usec


def sec_split_usec(ts):
    return int(ts), int((ts - int(ts)) * 1e6)


# a simple passive timer
class PassiveTimer(object):

    # timeout_sec = None means forever
    def __init__(self, timeout_sec):
        if timeout_sec is not None:
            self.expr_sec = time.time() + timeout_sec
        else:
            self.expr_sec = None

    def has_expired(self):
        # if no timeout was set - return always false
        if self.expr_sec is None:
            return False

        return(time.time() > self.expr_sec)


def is_valid_ipv4(addr):
    try:
        socket.inet_pton(socket.AF_INET, addr)
        return True
    except(socket.error, TypeError):
        return False


def is_valid_ipv6(addr):
    try:
        socket.inet_pton(socket.AF_INET6, addr)
        return True
    except(socket.error, TypeError):
        return False


def is_valid_mac(mac):
    return bool(re.match("[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                mac.lower()))


def list_remove_dup(l):
    tmp = list()

    for x in l:
        if x not in tmp:
            tmp.append(x)

    return tmp


def bitfield_to_list(bf):
    rc = []
    bitpos = 0

    while bf > 0:
        if bf & 0x1:
            rc.append(bitpos)
        bitpos += 1
        bf = bf >> 1

    return rc


def set_window_always_on_top(title):
    # we need the GDK module, if not available - ignroe this command
    try:
        if sys.version_info < (3, 0):
            from gtk import gdk
        else:
            # from gi.repository import Gdk as gdk
            return

    except ImportError:
        return

    # search the window and set it as above
    root = gdk.get_default_root_window()

    for id in root.property_get('_NET_CLIENT_LIST')[2]:
        w = gdk.window_foreign_new(id)
        if w:
            name = w.property_get('WM_NAME')[2]
            if title in name:
                w.set_keep_above(True)
                gdk.window_process_all_updates()
                break


def bitfield_to_str(bf):
    lst = bitfield_to_list(bf)
    return "-" if not lst else ', '.join([str(x) for x in lst])


# https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/
def natural_sorted_key(val):
    return [int(c) if c.isdigit() else c for c in re.split(r'(\d+)', val)]



#!/usr/bin/python3
# STLPktBuilder(pkt=Ether()/IP())
import base64
import copy
import imp
import json
import os
import sys
import traceback

import yaml

import scapy
from scapy.all import Raw
from scapy.error import Scapy_Exception
from scapy.layers.inet import IP
from scapy.layers.l2 import Dot3, Ether
from scapy.utils import RawPcapReader, ltoa, rdpcap

from trex_stl_lib.trex_stl_exceptions import STLArgumentError, STLError
from trex_stl_lib.trex_stl_packet_builder_interface import CTrexPktBuilderInterface
from trex_stl_lib.trex_stl_packet_builder_scapy import STLPktBuilder
from trex_stl_lib.trex_stl_types import OrderedDict, validate_type, verify_exclusive_arg
from trex_stl_lib.utils.text_opts import format_num


# base class for TX mode
class STLTXMode(object):
    """ mode rate speed """

    def __init__(self, pps=None, bps_L1=None, bps_L2=None, percentage=None):
        """
        Speed can be given in packets per second(pps), L2/L1 bps, or port percent
        Use only one unit.
        you can enter pps =10000 oe bps_L1=10

        :parameters:
            pps : float
               Packets per second

            bps_L1 : float
               Bits per second L1(with IPG)

            bps_L2 : float
               Bits per second L2(Ethernet-FCS)

            percentage : float
               Link interface percent(0-100). Example: 10 is 10% of the port link setup

        .. code-block:: python

            # STLTXMode Example

            mode = STLTXCont(pps = 10)

            mode = STLTXCont(bps_L1 = 10000000) #10mbps L1

            mode = STLTXCont(bps_L2 = 10000000) #10mbps L2

            mode = STLTXCont(percentage = 10)   #10%

        """

        args = [pps, bps_L1, bps_L2, percentage]

        # default
        if all([x is None for x in args]):
            pps = 1.0
        else:
            verify_exclusive_arg(args)

        self.fields = {'rate': {}}

        if pps is not None:
            validate_type('pps', pps, [float, int])

            self.fields['rate']['type'] = 'pps'
            self.fields['rate']['value'] = pps

        elif bps_L1 is not None:
            validate_type('bps_L1', bps_L1, [float, int])

            self.fields['rate']['type'] = 'bps_L1'
            self.fields['rate']['value'] = bps_L1

        elif bps_L2 is not None:
            validate_type('bps_L2', bps_L2, [float, int])

            self.fields['rate']['type'] = 'bps_L2'
            self.fields['rate']['value'] = bps_L2

        elif percentage is not None:
            validate_type('percentage', percentage, [float, int])
            if not(percentage > 0 and percentage <= 100):
                raise STLArgumentError('percentage', percentage)

            self.fields['rate']['type'] = 'percentage'
            self.fields['rate']['value'] = percentage

    def to_json(self):
        return self.fields


# continuous mode
class STLTXCont(STLTXMode):
    """ Continuous mode """

    def __init__(self, **kwargs):
        """
        Continuous mode

         see :class:`trex_stl_lib.trex_stl_streams.STLTXMode` for rate

        .. code-block:: python

            # STLTXCont Example

            mode = STLTXCont(pps = 10)

        """
        super(STLTXCont, self).__init__(**kwargs)

        self.fields['type'] = 'continuous'

    @staticmethod
    def __str__():
        return "Continuous"

# single burst mode


class STLTXSingleBurst(STLTXMode):
    """ Single burst mode """

    def __init__(self, total_pkts=1, **kwargs):
        """
        Single burst mode

            :parameters:
                 total_pkts : int
                    Number of packets for this burst

         see :class:`trex_stl_lib.trex_stl_streams.STLTXMode` for rate

        .. code-block:: python

            # STLTXSingleBurst Example

            mode = STLTXSingleBurst( pps = 10, total_pkts = 1)

        """

        if not isinstance(total_pkts, int):
            raise STLArgumentError('total_pkts', total_pkts)

        super(STLTXSingleBurst, self).__init__(**kwargs)

        self.fields['type'] = 'single_burst'
        self.fields['total_pkts'] = total_pkts

    @staticmethod
    def __str__():
        return "Single Burst"

# multi burst mode


class STLTXMultiBurst(STLTXMode):
    """ Multi-burst mode """

    def __init__(self,
                 pkts_per_burst=1,
                 ibg=0.0,   # usec not SEC
                 count=1,
                 **kwargs):
        """
        Multi-burst mode

        :parameters:

             pkts_per_burst: int
                Number of packets per burst

              ibg : float
                Inter-burst gap in usec 1,000,000.0 is 1 sec

              count : int
                Number of bursts

         see :class:`trex_stl_lib.trex_stl_streams.STLTXMode` for rate

        .. code-block:: python

            # STLTXMultiBurst Example

            mode = STLTXMultiBurst(pps = 10, pkts_per_burst = 1,count 10, ibg=10.0)

        """

        if not isinstance(pkts_per_burst, int):
            raise STLArgumentError('pkts_per_burst', pkts_per_burst)

        if not isinstance(ibg, (int, float)):
            raise STLArgumentError('ibg', ibg)

        if not isinstance(count, int):
            raise STLArgumentError('count', count)

        super(STLTXMultiBurst, self).__init__(**kwargs)

        self.fields['type'] = 'multi_burst'
        self.fields['pkts_per_burst'] = pkts_per_burst
        self.fields['ibg'] = ibg
        self.fields['count'] = count

    @staticmethod
    def __str__():
        return "Multi Burst"


STLStreamDstMAC_CFG_FILE = 0
STLStreamDstMAC_PKT = 1
STLStreamDstMAC_ARP = 2


class STLFlowStatsInterface(object):
    def __init__(self, pg_id):
        self.fields = {}
        self.fields['enabled'] = True
        self.fields['stream_id'] = pg_id

    def to_json(self):
        """ Dump as json"""
        return dict(self.fields)

    @staticmethod
    def defaults():
        return {'enabled': False}


class STLFlowStats(STLFlowStatsInterface):
    """ Define per stream basic stats

    .. code-block:: python

        # STLFlowStats Example

        flow_stats = STLFlowStats(pg_id = 7)

    """

    def __init__(self, pg_id):
        super(STLFlowStats, self).__init__(pg_id)
        self.fields['rule_type'] = 'stats'


class STLFlowLatencyStats(STLFlowStatsInterface):
    """ Define per stream basic stats + latency, jitter, packet reorder/loss

    .. code-block:: python

        # STLFlowLatencyStats Example

        flow_stats = STLFlowLatencyStats(pg_id = 7)

    """

    def __init__(self, pg_id):
        super(STLFlowLatencyStats, self).__init__(pg_id)
        self.fields['rule_type'] = 'latency'


class STLStream(object):
    """ One stream object. Includes mode, Field Engine mode packet template and Rx stats

        .. code-block:: python

            # STLStream Example


            base_pkt =  Ether()/IP(src="16.0.0.1",dst="48.0.0.1")/UDP(dport=12,sport=1025)
            pad = max(0, size - len(base_pkt)) * 'x'

            STLStream( isg = 10.0, # star in delay
                       name    ='S0',
                       packet = STLPktBuilder(pkt = base_pkt/pad),
                       mode = STLTXSingleBurst( pps = 10, total_pkts = 1),
                       next = 'S1'), # point to next stream


    """

    def __init__(self,
                 name=None,
                 packet=None,
                 mode=STLTXCont(pps=1),
                 enabled=True,
                 self_start=True,
                 isg=0.0,
                 flow_stats=None,
                 next=None,
                 stream_id=None,
                 action_count=0,
                 random_seed=0,
                 mac_src_override_by_pkt=None,
                 mac_dst_override_mode=None,  # see  STLStreamDstMAC_xx
                 dummy_stream=False
                 ):
        """
        Stream object

        :parameters:

                  name  : string
                       Name of the stream. Required if this stream is dependent on another stream,
                       and another stream needs to refer to this stream by name.

                  packet :  STLPktBuilder see :class:
                  `trex_stl_lib.trex_stl_packet_builder_scapy.STLPktBuilder`
                       Template packet and field engine program. Example:
                       packet = STLPktBuilder(pkt = base_pkt/pad)

                  mode :  :class:`trex_stl_lib.trex_stl_streams.STLTXCont` or :
                  class:`trex_stl_lib.trex_stl_streams.STLTXSingleBurst`  or  :
                      class:`trex_stl_lib.trex_stl_streams.STLTXMultiBurst`

                  enabled : bool
                      Indicates whether the stream is enabled.

                  self_start : bool
                      If False, another stream activates it.

                  isg : float
                     Inter-stream gap in usec. Time to wait until the stream
                     sends the first packet.

                  flow_stats : :class:`trex_stl_lib.trex_stl_streams.STLFlowStats`
                      Per stream statistic object. See: STLFlowStats

                  next : string
                      Name of the stream to activate.

                  stream_id :
                        For use by HLTAPI.

                  action_count : uint16_t
                        If there is a next stream, number of loops before stopping.
                        Default: 0(unlimited).

                  random_seed: uint16_t
                       If given, the seed for this stream will be this value.
                       Useful if you need a deterministic random value.

                  mac_src_override_by_pkt : bool
                        Template packet sets src MAC.

                  mac_dst_override_mode=None : STLStreamDstMAC_xx
                        Template packet sets dst MAC.

                  dummy_stream : bool
                        For delay purposes, will not be sent.
        """

        # type checking
        validate_type('mode', mode, STLTXMode)
        validate_type('packet', packet, (type(None), CTrexPktBuilderInterface))
        validate_type('flow_stats', flow_stats,
                      (type(None), STLFlowStatsInterface))
        validate_type('enabled', enabled, bool)
        validate_type('self_start', self_start, bool)
        validate_type('isg', isg, (int, float))
        validate_type('stream_id', stream_id, (type(None), int))
        validate_type('random_seed', random_seed, int)
        validate_type('dummy_stream', dummy_stream, bool)

        if(type(mode) == STLTXCont) and(next is not None):
            raise STLError("Continuous stream cannot have a next stream ID")

        # tag for the stream and next - can be anything
        self.name = name
        self.next = next

        # save for easy construct code from stream object
        self.mac_src_override_by_pkt = mac_src_override_by_pkt
        self.mac_dst_override_mode = mac_dst_override_mode
        self.id = stream_id

        self.fields = {}

        int_mac_src_override_by_pkt = 0
        int_mac_dst_override_mode = 0

        if mac_src_override_by_pkt is None:
            int_mac_src_override_by_pkt = 0
            if packet:
                if packet.is_default_src_mac() is False:
                    int_mac_src_override_by_pkt = 1

        else:
            int_mac_src_override_by_pkt = int(mac_src_override_by_pkt)

        if mac_dst_override_mode is None:
            int_mac_dst_override_mode = 0
            if packet:
                if packet.is_default_dst_mac() is False:
                    int_mac_dst_override_mode = STLStreamDstMAC_PKT
        else:
            int_mac_dst_override_mode = int(mac_dst_override_mode)

        self.is_default_mac = not(
            int_mac_src_override_by_pkt or int_mac_dst_override_mode)

        self.fields['flags'] = (int_mac_src_override_by_pkt & 1) + \
            ((int_mac_dst_override_mode & 3) << 1) + (int(dummy_stream) << 3)

        self.fields['action_count'] = action_count

        # basic fields
        self.fields['enabled'] = enabled
        self.fields['self_start'] = self_start
        self.fields['isg'] = isg

        if random_seed != 0:
            self.fields['random_seed'] = random_seed  # optional

        # mode
        self.fields['mode'] = mode.to_json()
        self.mode_desc = str(mode)

        # packet
        self.fields['packet'] = {}
        self.fields['vm'] = {}

        if not packet:
            packet = STLPktBuilder(pkt=Ether()/IP())
            if dummy_stream:
                self.packet_desc = 'Dummy'

        self.scapy_pkt_builder = packet
        # packet builder
        packet.compile()

        # packet and VM
        self.fields['packet'] = packet.dump_pkt()
        self.fields['vm'] = packet.get_vm_data()

        self.pkt = base64.b64decode(self.fields['packet']['binary'])

        # this is heavy, calculate lazy
        self.packet_desc = None

        if not flow_stats:
            self.fields['flow_stats'] = STLFlowStats.defaults()
        else:
            self.fields['flow_stats'] = flow_stats.to_json()

    def __str__(self):
        s = "Stream Name: {0}\n".format(self.name)
        s += "Stream Next: {0}\n".format(self.next)
        s += "Stream JSON:\n{0}\n".format(json.dumps(
            self.fields, indent=4, separators=(',', ': '), sort_keys=True))
        return s

    def to_json(self):
        """
        Return json format
        """
        return dict(self.fields)

    def get_id(self):
        """ Get the stream id after resolution  """
        return self.id

    def has_custom_mac_addr(self):
        """ Return True if src or dst MAC were set as custom """
        return not self.is_default_mac

    def get_name(self):
        """ Get the stream name """
        return self.name

    def get_next(self):
        """ Get next stream object """
        return getattr(self, '__next__', None)

    def has_flow_stats(self):
        """ Return True if stream was configured with flow stats """
        return self.fields['flow_stats']['enabled']

    def get_pkt(self):
        """ Get packet as string """
        return self.pkt

    def get_pkt_len(self, count_crc=True):
        """ Get packet number of bytes  """
        pkt_len = len(self.get_pkt())
        if count_crc:
            pkt_len += 4

        return pkt_len

    def get_pkt_type(self):
        """ Get packet description. Example: IP:UDP """
        if self.packet_desc is None:
            self.packet_desc = STLPktBuilder.pkt_layers_desc_from_buffer(
                self.get_pkt())

        return self.packet_desc

    def get_mode(self):
        return self.mode_desc

    @staticmethod
    def get_rate_from_field(rate_json):
        """ Get rate from json  """
        t = rate_json['type']
        v = rate_json['value']

        if t == "pps":
            return format_num(v, suffix="pps")
        elif t == "bps_L1":
            return format_num(v, suffix="bps(L1)")
        elif t == "bps_L2":
            return format_num(v, suffix="bps(L2)")
        elif t == "percentage":
            return format_num(v, suffix="%")

    def get_rate(self):
        return self.get_rate_from_field(self.fields['mode']['rate'])

    def to_pkt_dump(self):
        """ Print packet description from Scapy  """
        if self.name:
            print(("Stream Name: ", self.name))
        scapy_b = self.scapy_pkt_builder
        if scapy_b and isinstance(scapy_b, STLPktBuilder):
            scapy_b.to_pkt_dump()
        else:
            print("Nothing to dump")

    def to_yaml(self):
        """ Convert to YAML  """
        y = {}

        if self.name:
            y['name'] = self.name

        if self.__next__:
            y['next'] = self.__next__

        y['stream'] = copy.deepcopy(self.fields)

        # some shortcuts for YAML
        rate_type = self.fields['mode']['rate']['type']
        rate_value = self.fields['mode']['rate']['value']

        y['stream']['mode'][rate_type] = rate_value
        del y['stream']['mode']['rate']

        return y

    # returns the Python code(text) to build this stream,
    # inside the code it will be in variable "stream"
    def to_code(self):
        """ Convert to Python code as profile  """
        layer = Ether(self.pkt)
        layer.hide_defaults()          # remove fields with default values
        imports_arr = []
        layers_commands = []
        # remove checksums, add imports if needed
        while layer:
            layer_class = layer.__class__.__name__
            if layer_class not in vars(scapy.layers.all):  # custom import
                found_import = False
                for module_path, module in list(sys.modules.items()):
                    if not module_path.startswith(('scapy.layers', 'scapy.contrib')):
                        continue
                    import_string = 'from %s import %s' % (
                        module_path, layer_class)
                    if import_string in imports_arr:  # already present in extra imports
                        found_import = True
                        break
                    if hasattr(module, layer_class):  # add as extra import
                        imports_arr.append(import_string)
                        found_import = True
                        break
                if not found_import:
                    raise STLError(
                        'Could not determine import of layer %s' % layer.name)

            # remove checksums
            for chksum_name in('cksum', 'chksum'):
                if chksum_name in layer.fields:
                    del layer.fields[chksum_name]

            # remove Paddings(FCS etc.)
            if isinstance(layer, 'Padding'):
                break

            payload = layer.payload
            layer.remove_payload()

            if isinstance(layer, Raw):
                payload_data = bytes(layer)
                # compact form Raw('x' * 100) etc.
                if payload_data == payload_data[0:1] * len(payload_data):
                    layer_command = '%s * %s)' % (
                        Raw(payload_data[0:1]).command().rstrip(')'), len(payload_data))
                else:
                    layer_command = layer.command()
                layers_commands.append(layer_command)
                break

            layers_commands.append(layer.command())

            layer = payload

        imports = '\n'.join(imports_arr)
        packet_code = 'packet =(' + \
            (' / \n          ').join(layers_commands) + ')'

        vm_list = []
        for inst in self.fields['vm']['instructions']:
            if inst['type'] == 'flow_var':
                vm_list.append(
                    "STLVmFlowVar(name='{name}', size={size}, op='{op}', init_value={init_value},"
                    + "min_value={min_value}, max_value={max_value}, step={step})"
                    .format(**inst))
            elif inst['type'] == 'write_flow_var':
                vm_list.append(
                    "STLVmWrFlowVar(fv_name='{name}', pkt_offset={pkt_offset}, "
                    + "add_val={add_value}, is_big={is_big_endian})".format(**inst))
            elif inst['type'] == 'write_mask_flow_var':
                inst = copy.copy(inst)
                inst['mask'] = hex(inst['mask'])
                vm_list.append(
                    ("STLVmWrMaskFlowVar(fv_name='{name}', pkt_offset={pkt_offset}, "
                     + "pkt_cast_size={pkt_cast_size},"
                     + "mask={mask}, shift={shift}, add_value={add_value}, "
                     + "is_big={is_big_endian})").format(**inst))
            elif inst['type'] == 'fix_checksum_ipv4':
                vm_list.append(
                    "STLVmFixIpv4(offset={pkt_offset})".format(**inst))
            elif inst['type'] == 'trim_pkt_size':
                vm_list.append(
                    "STLVmTrimPktSize(fv_name='{name}')".format(**inst))
            elif inst['type'] == 'tuple_flow_var':
                inst = copy.copy(inst)
                inst['ip_min'] = ltoa(inst['ip_min'])
                inst['ip_max'] = ltoa(inst['ip_max'])
                vm_list.append(
                    ("STLVmTupleGen(name='{name}', ip_min='{ip_min}', "
                     + "ip_max='{ip_max}', port_min={port_min}, "
                     + "port_max={port_max}, limit_flows={limit_flows}, "
                     + "flags={flags})").format(**inst))
            elif inst['type'] == 'flow_var_rand_limit':
                vm_list.append(
                    "STLVmFlowVarRepetableRandom(name='{name}', size={size}, limit={limit}, "
                    + "seed={seed}, min_value={min_value}, max_value={max_value})"
                    .format(**inst))

        vm_code = 'vm = STLScVmRaw([' + \
            ',\n                 '.join(vm_list) + '])'
        stream_params_list = []
        stream_params_list.append(
            'packet = STLPktBuilder(pkt = packet, vm = vm)')
        if default_STLStream.name != self.name:
            stream_params_list.append('name = %s' %
                                      STLStream.__add_quotes(self.name))
        if default_STLStream.fields['enabled'] != self.fields['enabled']:
            stream_params_list.append('enabled = %s' % self.fields['enabled'])
        if default_STLStream.fields['self_start'] != self.fields['self_start']:
            stream_params_list.append('self_start = %s' %
                                      self.fields['self_start'])
        if default_STLStream.fields['isg'] != self.fields['isg']:
            stream_params_list.append('isg = %s' % self.fields['isg'])
        if default_STLStream.fields['flow_stats'] != self.fields['flow_stats']:
            stream_params_list.append(
                'flow_stats = STLFlowStats(%s)' % self.fields['flow_stats']['stream_id'])
        if default_STLStream.__next__ != self.__next__:
            stream_params_list.append('next = %s' %
                                      STLStream.__add_quotes(self.__next__))
        if default_STLStream.id != self.id:
            stream_params_list.append('stream_id = %s' % self.id)
        if default_STLStream.fields['action_count'] != self.fields['action_count']:
            stream_params_list.append(
                'action_count = %s' % self.fields['action_count'])
        if 'random_seed' in self.fields:
            stream_params_list.append(
                'random_seed = %s' % self.fields.get('random_seed', 0))
        if default_STLStream.mac_src_override_by_pkt != self.mac_src_override_by_pkt:
            stream_params_list.append(
                'mac_src_override_by_pkt = %s' % self.mac_src_override_by_pkt)
        if default_STLStream.mac_dst_override_mode != self.mac_dst_override_mode:
            stream_params_list.append(
                'mac_dst_override_mode = %s' % self.mac_dst_override_mode)

        mode_args = ''
        for key, value in list(self.fields['mode'].items()):
            if key not in('rate', 'type'):
                mode_args += '%s = %s, ' % (key, value)
        mode_args += '%s = %s' % (self.fields['mode']['rate']
                                  ['type'], self.fields['mode']['rate']['value'])
        if self.mode_desc == STLTXCont.__str__():
            stream_params_list.append('mode = STLTXCont(%s)' % mode_args)
        elif self.mode_desc == STLTXSingleBurst().__str__():
            stream_params_list.append(
                'mode = STLTXSingleBurst(%s)' % mode_args)
        elif self.mode_desc == STLTXMultiBurst().__str__():
            stream_params_list.append('mode = STLTXMultiBurst(%s)' % mode_args)
        else:
            raise STLError('Could not determine mode: %s' % self.mode_desc)

        stream = "stream = STLStream(" + \
            ',\n                   '.join(stream_params_list) + ')'
        return '\n'.join([imports, packet_code, vm_code, stream])

    # add quoted for string, or leave as is if other type
    @staticmethod
    def __add_quotes(arg):
        if type(arg) is str:
            return "'%s'" % arg
        return arg

    # used to replace non-printable characters with hex
    @staticmethod
    def __replchars_to_hex(match):
        return r'\x{0:02x}'.format(ord(match.group()))

    def dump_to_yaml(self, yaml_file=None):
        """ Print as yaml  """
        yaml_dump = yaml.dump([self.to_yaml()], default_flow_style=False)

        # write to file if provided
        if yaml_file:
            with open(yaml_file, 'w') as f:
                f.write(yaml_dump)

        return yaml_dump


class YAMLLoader(object):

    def __init__(self, yaml_file):
        self.yaml_path = os.path.dirname(yaml_file)
        self.yaml_file = yaml_file

    def __parse_packet(self, packet_dict):

        packet_type = set(packet_dict).intersection(['binary', 'pcap'])
        if len(packet_type) != 1:
            raise STLError(
                "Packet section must contain either 'binary' or 'pcap'")

        if 'binary' in packet_type:
            try:
                pkt_str = base64.b64decode(packet_dict['binary'])
            except TypeError:
                raise STLError("'binary' field is not a valid packet format")

            builder = STLPktBuilder(pkt_buffer=pkt_str)

        elif 'pcap' in packet_type:
            pcap = os.path.join(self.yaml_path, packet_dict['pcap'])

            if not os.path.exists(pcap):
                raise STLError("'pcap' - cannot find '{0}'".format(pcap))

            builder = STLPktBuilder(pkt=pcap)

        return builder

    def __parse_mode(self, mode_obj):
        if not mode_obj:
            return None

        rate_parser = set(mode_obj).intersection(
            ['pps', 'bps_L1', 'bps_L2', 'percentage'])
        if len(rate_parser) != 1:
            raise STLError(
                "'rate' must contain exactly one from 'pps', 'bps_L1', 'bps_L2', 'percentage'")

        rate_type = rate_parser.pop()
        rate = {rate_type: mode_obj[rate_type]}

        mode_type = mode_obj.get('type')

        if mode_type == 'continuous':
            mode = STLTXCont(**rate)

        elif mode_type == 'single_burst':
            defaults = STLTXSingleBurst()
            mode = STLTXSingleBurst(total_pkts=mode_obj.get('total_pkts',
                                                            defaults.fields['total_pkts']),
                                    **rate)

        elif mode_type == 'multi_burst':
            defaults = STLTXMultiBurst()
            mode = STLTXMultiBurst(pkts_per_burst=mode_obj.get('pkts_per_burst',
                                                               defaults.fields['pkts_per_burst']),
                                   ibg=mode_obj.get(
                                       'ibg', defaults.fields['ibg']),
                                   count=mode_obj.get(
                                       'count', defaults.fields['count']),
                                   **rate)

        else:
            raise STLError(
                "mode type can be 'continuous', 'single_burst' or 'multi_burst")

        return mode

    def __parse_flow_stats(self, flow_stats_obj):

        # no such object
        if not flow_stats_obj or flow_stats_obj.get('enabled') is False:
            return None

        pg_id = flow_stats_obj.get('stream_id')
        if pg_id is None:
            raise STLError(
                "Enabled RX stats section must contain 'stream_id' field")

        return STLFlowStats(pg_id=pg_id)

    def __parse_stream(self, yaml_object):
        s_obj = yaml_object['stream']

        # parse packet
        packet = s_obj.get('packet')
        if not packet:
            raise STLError("YAML file must contain 'packet' field")

        builder = self.__parse_packet(packet)

        # mode
        mode = self.__parse_mode(s_obj.get('mode'))

        # rx stats
        flow_stats = self.__parse_flow_stats(s_obj.get('flow_stats'))

        defaults = default_STLStream
        # create the stream
        stream = STLStream(name=yaml_object.get('name'),
                           packet=builder,
                           mode=mode,
                           flow_stats=flow_stats,
                           enabled=s_obj.get(
                               'enabled', defaults.fields['enabled']),
                           self_start=s_obj.get(
                               'self_start', defaults.fields['self_start']),
                           isg=s_obj.get('isg', defaults.fields['isg']),
                           next=yaml_object.get('next'),
                           action_count=s_obj.get(
                               'action_count', defaults.fields['action_count']),
                           mac_src_override_by_pkt=s_obj.get(
                               'mac_src_override_by_pkt', 0),
                           mac_dst_override_mode=s_obj.get(
                               'mac_src_override_by_pkt', 0)
                           )

        # hack the VM fields for now
        if 'vm' in s_obj:
            stream.fields['vm'].update(s_obj['vm'])

        return stream

    def parse(self):
        with open(self.yaml_file, 'r') as f:
            # read YAML and pass it down to stream object
            yaml_str = f.read()

            try:
                objects = yaml.safe_load(yaml_str)
            except yaml.parser.ParserError as e:
                raise STLError(str(e))

            streams = [self.__parse_stream(object) for object in objects]

            return streams


# profile class
class STLProfile(object):
    """ Describe a list of streams

        .. code-block:: python

            # STLProfile Example

            profile =  STLProfile( [ STLStream( isg = 10.0, # star in delay
                                        name    ='S0',
                                        packet = STLPktBuilder(pkt = base_pkt/pad),
                                        mode = STLTXSingleBurst( pps = 10,
                                        total_pkts = self.burst_size),
                                        next = 'S1'), # point to next stream

                             STLStream( self_start = False, # stream is
                                        disabled enable trow S0
                                        name    ='S1',
                                        packet  = STLPktBuilder(pkt = base_pkt1/pad),
                                        mode    = STLTXSingleBurst( pps = 10,
                                        total_pkts = self.burst_size),
                                        next    = 'S2' ),

                             STLStream(  self_start = False, # stream is  disabled enable trow S0
                                         name   ='S2',
                                         packet = STLPktBuilder(pkt = base_pkt2/pad),
                                         mode = STLTXSingleBurst( pps = 10,
                                         total_pkts = self.burst_size )
                                        )
                            ]).get_streams()



    """

    def __init__(self, streams=None):
        """

            :parameters:

                  streams  : list of :class:`trex_stl_lib.trex_stl_streams.STLStream`
                       a list of stream objects

        """

        if streams is None:
            streams = []

        if not type(streams) == list:
            streams = [streams]

        if not all([isinstance(stream, STLStream) for stream in streams]):
            raise STLArgumentError('streams', streams, valid_values=STLStream)

        self.streams = streams
        self.meta = None

    def get_streams(self):
        """ Get the list of streams"""
        return self.streams

    def __str__(self):
        return '\n'.join([str(stream) for stream in self.streams])

    def is_pauseable(self):
        return all([x.get_mode() == "Continuous" for x in self.get_streams()])

    def has_custom_mac_addr(self):
        return any([x.has_custom_mac_addr() for x in self.get_streams()])

    def has_flow_stats(self):
        return any([x.has_flow_stats() for x in self.get_streams()])

    @staticmethod
    def load_yaml(yaml_file):
        """ Load(from YAML file) a profile with a number of streams"""

        # check filename
        if not os.path.isfile(yaml_file):
            raise STLError("file '{0}' does not exists".format(yaml_file))

        yaml_loader = YAMLLoader(yaml_file)
        streams = yaml_loader.parse()

        profile = STLProfile(streams)
        profile.meta = {'type': 'yaml'}

        return profile

    @staticmethod
    def get_module_tunables(module):
        # remove self and variables
        func = module.register().get_streams
        argc = func.__code__.co_argcount
        tunables = func.__code__.co_varnames[1:argc]

        # fetch defaults
        defaults = func.__defaults__
        if defaults is None:
            return {}
        if len(defaults) != (argc - 1):
            raise STLError(
                "Module should provide default values for all arguments on get_streams()")

        output = {}
        for t, d in zip(tunables, defaults):
            output[t] = d

        return output

    @staticmethod
    def load_py(python_file, direction=0, port_id=0, **kwargs):
        """ Load from Python profile """

        # check filename
        if not os.path.isfile(python_file):
            raise STLError("File '{0}' does not exist".format(python_file))

        basedir = os.path.dirname(python_file)
        sys.path.insert(0, basedir)

        try:
            file = os.path.basename(python_file).split('.')[0]
            module = __import__(file, globals(), locals(), [], 0)
            imp.reload(module)  # reload the update

            t = STLProfile.get_module_tunables(module)
            # for arg in kwargs:
            #    if not arg in t:
            #        raise STLError("Profile {0} does not support tunable '{1}'
            #  - supported tunables are: '{2}'".format(python_file, arg, t))

            streams = module.register().get_streams(direction=direction,
                                                    port_id=port_id,
                                                    **kwargs)
            profile = STLProfile(streams)

            profile.meta = {'type': 'python',
                            'tunables': t}

            return profile

        except Exception:
            a, b, tb = sys.exc_info()
            x = ''.join(traceback.format_list(traceback.extract_tb(
                tb)[1:])) + a.__name__ + ": " + str(b) + "\n"

            summary = "\nPython Traceback follows:\n\n" + x
            raise STLError(summary)

        finally:
            sys.path.remove(basedir)

    # loop_count = 0 means loop forever

    @staticmethod
    def load_pcap(pcap_file,
                  ipg_usec=None,
                  speedup=1.0,
                  loop_count=1,
                  vm=None,
                  packet_hook=None,
                  split_mode=None,
                  min_ipg_usec=None):
        """ Convert a pcap file with a number of packets to a list of connected streams.

        packet1->packet2->packet3 etc

                :parameters:

                  pcap_file  : string
                       Name of the pcap file

                  ipg_usec   : float
                       Inter packet gap in usec. If IPG is None, IPG is taken from pcap file

                  speedup   : float
                       When reading the pcap file, divide IPG by this "speedup" factor.
                       Resulting IPG is sped up by this factor.

                  loop_count : uint16_t
                       Number of loops to repeat the pcap file

                  vm        :  list
                        List of Field engine instructions

                  packet_hook : Callable or function
                        will be applied to every packet

                  split_mode : str
                        should this PCAP be split to two profiles based on IPs / MACs
                        used for dual mode
                        can be 'MAC' or 'IP'

                  min_ipg_usec   : float
                       Minumum inter packet gap in usec. Used to guard from too small IPGs.

                 :return: STLProfile

        """

        if speedup <= 0:
            raise STLError('Speedup should be positive.')
        if min_ipg_usec and min_ipg_usec < 0:
            raise STLError('min_ipg_usec should not be negative.')

        # make sure IPG is not less than 0.001 usec
        if(ipg_usec is not None and(ipg_usec < 0.001 * speedup) and
           (min_ipg_usec is None or min_ipg_usec < 0.001)):
            raise STLError(
                "ipg_usec cannot be less than 0.001 usec: '{0}'".format(ipg_usec))

        if loop_count < 0:
            raise STLError("'loop_count' cannot be negative")

        try:
            if split_mode is None:
                pkts = PCAPReader(pcap_file).read_all(
                    ipg_usec, min_ipg_usec, speedup)
                if len(pkts) == 0:
                    raise STLError(
                        "'{0}' does not contain any packets".format(pcap_file))

                return STLProfile.__pkts_to_streams(pkts,
                                                    loop_count,
                                                    vm,
                                                    packet_hook)
            else:
                pkts_a, pkts_b = PCAPReader(pcap_file).read_all(
                    ipg_usec, min_ipg_usec, speedup, split_mode=split_mode)
                if not(pkts_a or pkts_b):
                    raise STLError(
                        "'%s' does not contain any packets." % pcap_file)
                elif not(pkts_a and pkts_b):
                    raise STLError(
                        "'%s' contains only one direction." % pcap_file)

                # swap is ts of first packet in b is earlier
                start_time_a = pkts_a[0][1]
                start_time_b = pkts_b[0][1]
                if start_time_b < start_time_a:
                    pkts_a, pkts_b = pkts_b, pkts_a

                # get last ts
                end_time_a = pkts_a[-1][1]
                end_time_b = pkts_b[-1][1]
                start_delay_usec = 1000
                if ipg_usec:
                    start_delay_usec = ipg_usec / speedup
                if min_ipg_usec and min_ipg_usec > start_delay_usec:
                    start_delay_usec = min_ipg_usec
                end_time = max(end_time_a, end_time_b)

                profile_a = STLProfile.__pkts_to_streams(pkts_a,
                                                         loop_count,
                                                         vm,
                                                         packet_hook,
                                                         start_delay_usec,
                                                         end_delay_usec=end_time - end_time_a)

                profile_b = STLProfile.__pkts_to_streams(pkts_b,
                                                         loop_count,
                                                         vm,
                                                         packet_hook,
                                                         start_delay_usec,
                                                         end_delay_usec=end_time - end_time_b)

                return profile_a, profile_b

        except Scapy_Exception as e:
            raise STLError(
                "failed to open PCAP file {0}: '{1}'".format(pcap_file, str(e)))

    @staticmethod
    def __pkts_to_streams(pkts, loop_count, vm, packet_hook, start_delay_usec=0, end_delay_usec=0):
        streams = []
        if packet_hook:
            pkts = [(packet_hook(cap), meta) for(cap, meta) in pkts]

        last_ts = 0
        for i, (cap, ts) in enumerate(pkts, start=1):
            isg = ts - last_ts
            last_ts = ts

            # handle last packet
            if i == len(pkts):
                if end_delay_usec:
                    next = 'delay_stream'
                    action_count = 0
                    streams.append(STLStream(name='delay_stream',
                                             mode=STLTXSingleBurst(
                                                 total_pkts=1, percentage=100),
                                             self_start=False,
                                             isg=end_delay_usec,
                                             action_count=loop_count,
                                             dummy_stream=True,
                                             next=1))
                else:
                    next = 1
                    action_count = loop_count
            else:
                next = i + 1
                action_count = 0

            if i == 1:
                streams.append(STLStream(name=1,
                                         packet=STLPktBuilder(
                                             pkt_buffer=cap, vm=vm),
                                         mode=STLTXSingleBurst(
                                             total_pkts=1, percentage=100),
                                         self_start=True,
                                         isg=isg + start_delay_usec,  # usec
                                         action_count=action_count,
                                         next=next))
            else:
                streams.append(STLStream(name=i,
                                         packet=STLPktBuilder(
                                             pkt_buffer=cap, vm=vm),
                                         mode=STLTXSingleBurst(
                                             total_pkts=1, percentage=100),
                                         self_start=False,
                                         isg=isg,  # usec
                                         action_count=action_count,
                                         next=next))

        profile = STLProfile(streams)
        profile.meta = {'type': 'pcap'}

        return profile

    @staticmethod
    def load(filename, direction=0, port_id=0, **kwargs):
        """ Load a profile by its type. Supported types are:
           * py
           * yaml
           * pcap file that converted to profile automaticly

           :Parameters:
              filename  : string as filename
              direction : profile's direction(if supported by the profile)
              port_id   : which port ID this profile is being loaded to
              kwargs    : forward those key-value pairs to the profile

        """

        x = os.path.basename(filename).split('.')
        suffix = x[1] if(len(x) == 2) else None

        if suffix == 'py':
            profile = STLProfile.load_py(
                filename, direction, port_id, **kwargs)

        elif suffix == 'yaml':
            profile = STLProfile.load_yaml(filename)

        elif suffix in ['cap', 'pcap']:
            profile = STLProfile.load_pcap(filename, speedup=1, ipg_usec=1e6)

        else:
            raise STLError("unknown profile file type: '{0}'".format(suffix))

        profile.meta['stream_count'] = len(profile.get_streams()) if isinstance(
            profile.get_streams(), list) else 1
        return profile

    @staticmethod
    def get_info(filename):
        profile = STLProfile.load(filename)
        return profile.meta

    def dump_as_pkt(self):
        """ Dump the profile as Scapy packet. If the packet is raw,
            convert it to Scapy before dumping it."""
        cnt = 0
        for stream in self.streams:
            print("=======================")
            print(("Stream %d" % cnt))
            print("=======================")
            cnt = cnt + 1
            stream.to_pkt_dump()

    def dump_to_yaml(self, yaml_file=None):
        """ Convert the profile to yaml """
        yaml_list = [stream.to_yaml() for stream in self.streams]
        yaml_str = yaml.dump(yaml_list, default_flow_style=False)

        # write to file if provided
        if yaml_file:
            with open(yaml_file, 'w') as f:
                f.write(yaml_str)

        return yaml_str

    def dump_to_code(self, profile_file=None):
        """ Convert the profile to Python native profile. """
        profile_dump = '''# !!! Auto-generated code !!!
from trex_stl_lib.api import *

class STLS1(object):
    def get_streams(self, direction = 0, **kwargs):
        streams = []
'''
        for stream in self.streams:
            profile_dump += ' '*8 + stream.to_code().replace('\n', '\n' + ' '*8) + '\n'
            profile_dump += ' '*8 + 'streams.append(stream)\n'
        profile_dump += '''
        return streams

def register():
    return STLS1()
'''
        # write to file if provided
        if profile_file:
            with open(profile_file, 'w') as f:
                f.write(profile_dump)

        return profile_dump

    def __len__(self):
        return len(self.streams)


class PCAPReader(object):
    def __init__(self, pcap_file):
        if not os.path.isfile(pcap_file):
            raise STLError("File '{0}' does not exist.".format(pcap_file))
        self.pcap_file = pcap_file

    def read_all(self, ipg_usec, min_ipg_usec, speedup, split_mode=None):
        # get the packets
        if split_mode is None:
            pkts = RawPcapReader(self.pcap_file).read_all()
        else:
            pkts = rdpcap(self.pcap_file)

        if not pkts:
            raise STLError("'%s' does not contain any packets." %
                           self.pcap_file)

        self.pkts_arr = []
        last_ts = 0
        # fix times
        for pkt in pkts:
            if split_mode is None:
                pkt_data, meta = pkt
                ts_usec = meta[0] * 1e6 + meta[1]
            else:
                pkt_data = pkt
                ts_usec = pkt.time * 1e6

            if ipg_usec is None:
                loco = locals()
                if 'prev_time' in loco:
                    delta_usec = (ts_usec - loco['prev_time']) / float(speedup)
                else:
                    delta_usec = 0
                if min_ipg_usec and delta_usec < min_ipg_usec:
                    delta_usec = min_ipg_usec
                prev_time = ts_usec
                last_ts += delta_usec
            else:  # user specified ipg
                if min_ipg_usec:
                    last_ts += min_ipg_usec
                elif ipg_usec:
                    last_ts += ipg_usec / float(speedup)
                else:
                    raise STLError(
                        'Please specify either min_ipg_usec or ipg_usec, not both.')
            self.pkts_arr.append([pkt_data, last_ts])

        if split_mode is None:
            return self.pkts_arr

        # we need to split
        self.graph = Graph()

        self.pkt_groups = [[], []]

        if split_mode == 'MAC':
            self.generate_mac_groups()
        elif split_mode == 'IP':
            self.generate_ip_groups()
        else:
            raise STLError('unknown split mode for PCAP')

        return self.pkt_groups

    # generate two groups based on MACs

    def generate_mac_groups(self):
        for i, (pkt, _) in enumerate(self.pkts_arr):
            if not isinstance(pkt, (Ether, Dot3)):
                raise STLError(
                    "Packet #{0} has an unknown L2 format: {1}".format(i, type(pkt)))
            self.graph.add(pkt.src, pkt.dst)

        # split the graph to two groups
        mac_groups = self.graph.split()

        for pkt, ts in self.pkts_arr:
            group = 1 if pkt.src in mac_groups[1] else 0
            self.pkt_groups[group].append((bytes(pkt), ts))

    # generate two groups based on IPs

    def generate_ip_groups(self):
        for i, (pkt, t) in enumerate(self.pkts_arr):
            if not isinstance(pkt, (Ether, Dot3)):
                raise STLError(
                    "Packet #{0} has an unknown L2 format: {1}".format(i, type(pkt)))
            ip = pkt.getlayer('IP')
            if not ip:
                ip = pkt.getlayer('IPv6')
            if not ip:
                continue
            self.graph.add(ip.src, ip.dst)

        # split the graph to two groups
        ip_groups = self.graph.split()

        for pkt, ts in self.pkts_arr:
            ip = pkt.getlayer('IP')
            if not ip:
                ip = pkt.getlayer('IPv6')
            group = 0
            if ip and ip.src in ip_groups[1]:
                group = 1
            self.pkt_groups[group].append((bytes(pkt), ts))


# a simple graph object - used to split to two groups
class Graph(object):
    def __init__(self):
        self.db = OrderedDict()
        self.debug = False

    def log(self, msg):
        if self.debug:
            print(msg)

    # add a connection v1 --> v2
    def add(self, v1, v2):
        # init value for v1
        if v1 not in self.db:
            self.db[v1] = set()

        # init value for v2
        if v2 not in self.db:
            self.db[v2] = set()

        # ignore self to self edges
        if v1 == v2:
            return

        # undirected - add two ways
        self.db[v1].add(v2)
        self.db[v2].add(v1)

    # create a 2-color of the graph if possible

    def split(self):
        color_a = set()
        color_b = set()

        # start with all
        nodes = list(self.db.keys())

        # process one by one
        while len(nodes) > 0:
            node = nodes.pop(0)

            friends = self.db[node]

            # node has never been seen - move to color_a
            if node not in color_a and node not in color_b:
                self.log("<NEW> {0} --> A".format(node))
                color_a.add(node)

            # node color
            node_color, other_color = (
                color_a, color_b) if node in color_a else(color_b, color_a)

            # check that the coloring is possible
            bad_friends = friends.intersection(node_color)
            if bad_friends:
                raise STLError(
                    "ERROR: failed to split PCAP file - {0} and {1} are in the same group"
                    .format(node, bad_friends))

            # add all the friends to the other color
            for friend in friends:
                self.log("<FRIEND> {0} --> {1}".format(friend,
                                                       'A' if other_color is color_a else 'B'))
                other_color.add(friend)

        return color_a, color_b


default_STLStream = STLStream()



#!/usr/bin/python3

import time

# a generic abstract class for implementing RX services using the server


class RXServiceAPI(object):

    # specify for which layer this service is
    LAYER_MODE_ANY = 0
    LAYER_MODE_L2 = 1
    LAYER_MODE_L3 = 2

    def __init__(self, port, layer_mode=LAYER_MODE_ANY, queue_size=100, timeout=None, retries=None, retry_delay=0.1):
        self.port = port
        self.queue_size = queue_size
        self.layer_mode = layer_mode
        self.timeout = timeout
        self.retries = retries
        if retries is None and timeout is None:
            self.retries = 0
        self.retry_delay = retry_delay
        self.init_ts = time.time()

    ################### virtual methods ######################

    def get_name(self):
        """
            returns the name of the service

            :returns:
                str

        """

        raise NotImplementedError()

    def pre_execute(self):
        """
            specific class code called before executing

            :returns:
                RC object

        """
        raise NotImplementedError()

    def generate_request(self):
        """
            generate a request to be sent to the server

            :returns:
                list of streams

        """
        raise NotImplementedError()

    def on_pkt_rx(self, pkt, start_ts):
        """
            called for each packet arriving on RX

            :parameters:
                'pkt' - the packet received
                'start_ts' - the time recorded when 'start' was called 

            :returns:
                None for fetching more packets
                RC object for terminating



        """
        raise NotImplementedError()

    def on_timeout(self):
        """
            called when a timeout occurs

            :returns:
                RC object

        """
        raise NotImplementedError()

    ##################### API ######################

    def execute(self, *a, **k):

        # sanity check
        rc = self.__sanity()
        if not rc:
            return rc

        # first cleanup
        rc = self.port.remove_all_streams()
        if not rc:
            return rc

        # start the iteration
        try:

            # add the stream(s)
            self.port.add_streams(self.generate_request(*a, **k))
            rc = self.port.set_rx_queue(size=self.queue_size)
            if not rc:
                return rc

            return self.__execute_internal()

        finally:
            # best effort restore
            self.port.remove_rx_queue()
            self.port.remove_all_streams()

    ##################### Internal ######################

    def __sanity(self):
        if not self.port.is_service_mode_on():
            return self.port.err('port service mode must be enabled for performing {0}. Please enable service mode'.format(self.get_name()))

        if self.layer_mode == RXServiceAPI.LAYER_MODE_L2:
            if not self.port.is_l2_mode():
                return self.port.err('{0} - requires L2 mode configuration'.format(self.get_name()))

        elif self.layer_mode == RXServiceAPI.LAYER_MODE_L3:
            if not self.port.is_l3_mode():
                return self.port.err('{0} - requires L3 mode configuration'.format(self.get_name()))

        # sanity
        if self.port.is_active():
            return self.port.err('{0} - port is active, please stop the port before executing command'.format(self.get_name()))

        # call the specific class implementation
        rc = self.pre_execute()
        if not rc:
            return rc

        return True

    # main resolve function

    def __execute_internal(self):

        # retry for 'retries' or until timeout
        index = 0
        while True:
            rc = self.execute_iteration()
            if rc is not None:
                return rc

            if(self.retries is not None and index >= self.retries or
               self.timeout is not None and time.time() - self.init_ts >= self.timeout):
                return self.on_timeout()

            index += 1
            time.sleep(self.retry_delay)

    def execute_iteration(self):

        mult = {'op': 'abs', 'type': 'percentage', 'value': 100}
        rc = self.port.start(mul=mult, force=False,
                             duration=-1, mask=0xffffffff)
        if not rc:
            return rc

        # save the start timestamp
        self.start_ts = rc.data()['ts']

        # block until traffic finishes
        while self.port.is_active():
            time.sleep(0.01)

        return self.wait_for_rx_response()

    def wait_for_rx_response(self):

        # we try to fetch response for 5 times
        polling = 5

        while polling > 0:

            # fetch the queue
            rx_pkts = self.port.get_rx_queue_pkts()

            # might be an error
            if not rx_pkts:
                return rx_pkts

            # for each packet - examine it
            for pkt in rx_pkts.data():
                rc = self.on_pkt_rx(pkt, self.start_ts)
                if rc is not None:
                    return rc

            if polling == 0:
                return None

            polling -= 1
            time.sleep(0.1)



#!/usr/bin/python3

import struct
from collections import OrderedDict, namedtuple

from scapy.layers.dhcp import BOOTP, DHCP
from scapy.layers.inet import IP, UDP
from scapy.layers.l2 import Ether
from scapy.packet import NoPayload

from .trex_stl_service_fast_parser import FastParser


class DHCPParser(FastParser):

    # message types
    DISCOVER = 1
    OFFER = 2
    ACK = 5
    NACK = 6
    RELEASE = 7

    def __init__(self):
        base_pkt = Ether(dst="ff:ff:ff:ff:ff:ff")/IP(src="0.0.0.0", dst="255.255.255.255")/UDP(sport=68, dport=67, chksum=0) \
            / BOOTP(chaddr=b'123456', xid=55555, yiaddr='1.2.3.4')/DHCP(options=[("message-type", "discover"), ("server_id", '1.2.3.4'), "end"])

        FastParser.__init__(self, base_pkt)

        self.add_field('Ethernet.dst', 'dstmac')
        self.add_field('Ethernet.src', 'srcmac')

        self.add_field('IP.ihl',      'ihl', fmt="!B")
        self.add_field('IP.dst',      'dstip')
        self.add_field('IP.src',      'srcip')
        self.add_field('IP.chksum',   'chksum')

        self.add_field('BOOTP.xid', 'xid')
        self.add_field('BOOTP.chaddr', 'chaddr')
        self.add_field('BOOTP.ciaddr', 'ciaddr')
        self.add_field('BOOTP.yiaddr', 'yiaddr')
        self.add_field('DHCP options.options', 'options',
                       getter=self.get_options, setter=self.set_options)

        msg_types = [{'id': 1, 'name': 'discover'},
                     {'id': 2, 'name': 'offer'},
                     {'id': 3, 'name': 'request'},
                     ]

        self.msg_types = {}

        for t in msg_types:
            self.msg_types[t['id']] = t
            self.msg_types[t['name']] = t

        opts = [{'id': 53, 'name': 'message-type',   'type': 'byte'},
                {'id': 54, 'name': 'server_id',      'type': 'int'},
                {'id': 50, 'name': 'requested_addr', 'type': 'int'},
                {'id': 51, 'name': 'lease-time',     'type': 'int'},
                {'id': 58, 'name': 'renewal_time',   'type': 'int'},
                {'id': 59, 'name': 'rebinding_time', 'type': 'int'},
                {'id': 1,  'name': 'subnet_mask',    'type': 'int'},
                {'id': 15, 'name': 'domain',         'type': 'str'},
                ]

        self.opts = {}

        for opt in opts:
            self.opts[opt['id']] = opt
            self.opts[opt['name']] = opt

    def get_options(self, pkt_bytes, info):

        # min length
        if len(pkt_bytes) < info['offset']:
            return None

        options = pkt_bytes[info['offset']:]

        opt = OrderedDict()
        index = 0
        while index < len(options):

            o = ord(options[index])
            index += 1

            # end
            if o == 255:
                break

            # pad
            elif o == 0:
                continue

            # fetch length
            olen = ord(options[index])
            index += 1

            # message type
            if o in self.opts:
                ot = self.opts[o]
                if ot['type'] == 'byte':
                    opt[ot['name']] = struct.unpack_from(
                        '!B', options, index)[0]

                elif ot['type'] == 'int':
                    opt[ot['name']] = struct.unpack_from(
                        '!I', options, index)[0]

                elif ot['type'] == 'str':
                    opt[ot['name']] = struct.unpack_from(
                        '!{0}s'.format(olen), options, index)[0]

                else:
                    raise Exception('unknown type: {0}'.format(ot['type']))

            else:
                raise Exception('unknown option: {0}'.format(o))

            # advance
            index += olen

        return opt

    def set_options(self, pkt_bytes, info, options):

        output = bytes()

        for o, v in list(options.items()):
            if o in self.opts:
                ot = self.opts[o]

                # write tag
                output += struct.pack('!B', ot['id'])

                # write the size and value
                if ot['type'] == 'byte':
                    output += struct.pack('!B', 1)
                    output += struct.pack('!B', v)

                elif ot['type'] == 'int':
                    output += struct.pack('!B', 4)
                    output += struct.pack('!I', v)

                elif ot['type'] == 'str':
                    output += struct.pack('!B', len(v))
                    output += struct.pack('!{0}s'.format(len(v)), v)

        # write end
        output += struct.pack('!B', 255)

        return pkt_bytes[:info['offset']] + output + pkt_bytes[info['offset'] + len(output):]

    def disc(self, xid, chaddr):
        '''
            generates a DHCP discovery packet
        '''

        # generate a new packet
        obj = self.clone()

        obj.options = {'message-type': 1}
        obj.xid = xid
        obj.chaddr = chaddr

        return obj.raw()

    def req(self, xid, chaddr, yiaddr):
        '''
            generate a new request packet
        '''

        # generate a new packet
        obj = self.clone()

        obj.options = {'message-type': 3, 'requested_addr': yiaddr}
        obj.xid = xid
        obj.chaddr = chaddr

        return obj.raw()

    def release(self, xid, client_mac, client_ip, server_mac, server_ip):
        '''
            generate a release request packet
        '''

        # generate a new packet
        obj = self.clone()

        obj.dstmac = server_mac
        obj.srcmac = client_mac

        obj.dstip = server_ip
        obj.srcip = client_ip

        obj.fix_chksum()

        obj.ciaddr = client_ip
        obj.chaddr = client_mac
        obj.xid = xid

        obj.options = {'message-type': 7, 'server_id': server_ip}

        return obj.raw()



#!/usr/bin/python3


def shallow_copy(x):
    return type(x)(x)


class ToggleFilter(object):
    """
    This class provides a "sticky" filter, that works by "toggling" items
    of the original database on and off.
    """

    def __init__(self, db_ref, show_by_default=True):
        """
        Instantiate a ToggleFilter object

        :parameters:
             db_ref : iterable
                an iterable object(i.e. list, set etc) that would serve as
                the reference db of the instance.
                Changes in that object will affect the output of
                ToggleFilter instance.

             show_by_default: bool
                decide if by default all the items are "on", i.e. these
                items will be presented if no other
                toggling occurred.

                default value : **True**

        """
        self._data = db_ref
        self._toggle_db = set()
        self._filter_method = filter
        self.__set_initial_state(show_by_default)

    def reset(self):
        """
        Toggles off all the items
        """
        self._toggle_db = set()

    def toggle_item(self, item_key):
        """
        Toggle a single item in/out.

        :parameters:
             item_key :
                an item the by its value the filter can decide to toggle
                or not.
                Example: int, str and so on.

        :return:
            + **True** if item toggled **into** the filtered items
            + **False** if item toggled **out from** the filtered items

        :raises:
            + KeyError, in case if item key is not part of the toggled
            list and not part of the referenced db.

        """
        if item_key in self._toggle_db:
            self._toggle_db.remove(item_key)
            return False
        elif item_key in self._data:
            self._toggle_db.add(item_key)
            return True
        else:
            raise KeyError("Provided item key isn't a key " +
                           "of the referenced data structure.")

    def toggle_items(self, *args):
        """
        Toggle multiple items in/out with a single call. Each item will be ha.

        :parameters:
             args : iterable
                an iterable object containing all item keys to be
                toggled in/out

        :return:
            + **True** if all toggled items were toggled **into**
            the filtered items
            + **False** if at least one of the items was toggled
            **out from** the filtered items

        :raises:
            + KeyError, in case if ont of the item keys was not part of
            the toggled list and not part of the referenced db.

        """
        # in python 3, 'map' returns an iterator, so wrapping with 'list'
        # call creates same effect for both python 2 and 3
        return all(list(map(self.toggle_item, args)))

    def filter_items(self):
        """
        Filters the pointed database by showing only the
        items mapped at toggle_db set.

        :returns:
            Filtered data of the original object.

        """
        return self._filter_method(self.__toggle_filter, self._data)

    # private methods

    def __set_initial_state(self, show_by_default):
        try:
            _state = (x for x in self._data)
            if isinstance(self._data, dict):
                self._filter_method = ToggleFilter.dict_filter
                if show_by_default:
                    self._toggle_db = set(self._data.keys())
                return
            elif isinstance(self._data, list):
                self._filter_method = ToggleFilter.list_filter
            elif isinstance(self._data, set):
                self._filter_method = ToggleFilter.set_filter
            elif isinstance(self._data, tuple):
                self._filter_method = ToggleFilter.tuple_filter
            if show_by_default:
                # assuming all relevant data with unique identifier
                self._toggle_db = set(shallow_copy(self._data))
            return _state
        except TypeError:
            raise TypeError("provided data object is not iterable")

    def __toggle_filter(self, x):
        return(x in self._toggle_db)

    # static utility methods

    @staticmethod
    def dict_filter(function, iterable):
        assert isinstance(iterable, dict)
        return {k: v
                for k, v in list(iterable.items())
                if function(k)}

    @staticmethod
    def list_filter(function, iterable):
        # in python 3, filter returns an iterator, so wrapping with
        # list creates same effect for both python 2 and 3
        return list(filter(function, iterable))

    @staticmethod
    def set_filter(function, iterable):
        return {x
                for x in iterable
                if function(x)}

    @staticmethod
    def tuple_filter(function, iterable):
        return tuple(filter(function, iterable))


if __name__ == "__main__":
    pass



#!/usr/bin/python3

import copy
import sys
import threading
import time
from collections import OrderedDict, deque, namedtuple

from trex_stl_lib.trex_stl_exceptions import STLError
from trex_stl_lib.trex_stl_types import StatNotAvailable, is_integer
from trex_stl_lib.utils import text_tables
from trex_stl_lib.utils.text_opts import (format_num, format_text,
                                          format_threshold)

GLOBAL_STATS = 'g'
PORT_STATS = 'p'
PORT_GRAPH = 'pg'
PORT_STATUS = 'ps'
STREAMS_STATS = 's'
LATENCY_STATS = 'ls'
LATENCY_HISTOGRAM = 'lh'
CPU_STATS = 'c'
MBUF_STATS = 'm'
EXTENDED_STATS = 'x'
EXTENDED_INC_ZERO_STATS = 'xz'

ALL_STATS_OPTS = [GLOBAL_STATS, PORT_STATS, PORT_STATUS, STREAMS_STATS, LATENCY_STATS,
                  PORT_GRAPH, LATENCY_HISTOGRAM, CPU_STATS, MBUF_STATS, EXTENDED_STATS,
                  EXTENDED_INC_ZERO_STATS]
COMPACT = [GLOBAL_STATS, PORT_STATS]
GRAPH_PORT_COMPACT = [GLOBAL_STATS, PORT_GRAPH]
SS_COMPAT = [GLOBAL_STATS, STREAMS_STATS]  # stream stats
LS_COMPAT = [GLOBAL_STATS, LATENCY_STATS]  # latency stats
LH_COMPAT = [GLOBAL_STATS, LATENCY_HISTOGRAM]  # latency histogram
UT_COMPAT = [GLOBAL_STATS, CPU_STATS, MBUF_STATS]  # utilization

ExportableStats = namedtuple('ExportableStats', ['raw_data', 'text_table'])


def round_float(f):
    return float("%.2f" % f) if type(f) is float else f


def try_int(i):
    try:
        return int(i)
    except Exception:
        return i

# deep mrege of dicts dst = src + dst


def deep_merge_dicts(dst, src):
    for k, v in list(src.items()):
        # if not exists - deep copy it
        if k not in dst:
            dst[k] = copy.deepcopy(v)
        else:
            if isinstance(v, dict):
                deep_merge_dicts(dst[k], v)

# BPS L1 from pps and BPS L2


def calc_bps_L1(bps, pps):
    if(pps == 0) or(bps == 0):
        return 0

    factor = bps / (pps * 8.0)
    return bps * (1 + (20 / factor))
#


def is_intable(value):
    try:
        int(value)
        return True
    except ValueError:
        return False

# use to calculate diffs relative to the previous values
# for example, BW


def calculate_diff(samples):
    total = 0.0

    weight_step = 1.0 / sum(range(0, len(samples)))
    weight = weight_step

    for i in range(0, len(samples) - 1):
        current = samples[i] if samples[i] > 0 else 1
        next = samples[i + 1] if samples[i + 1] > 0 else 1

        s = 100 * ((float(next) / current) - 1.0)

        # block change by 100%
        total += (min(s, 100) * weight)
        weight += weight_step

    return total


# calculate by absolute values and not relatives(useful for CPU usage in % and etc.)
def calculate_diff_raw(samples):
    total = 0.0

    weight_step = 1.0 / sum(range(0, len(samples)))
    weight = weight_step

    for i in range(0, len(samples) - 1):
        current = samples[i]
        next = samples[i + 1]

        total += ((next - current) * weight)
        weight += weight_step

    return total


get_number_of_bytes_cache = {}
# get number of bytes: '64b'->64, '9kb'->9000 etc.


def get_number_of_bytes(val):
    if val not in get_number_of_bytes_cache:
        get_number_of_bytes_cache[val] = int(val[:-1].replace('k', '000'))
    return get_number_of_bytes_cache[val]

# a simple object to keep a watch over a field


class WatchedField(object):

    def __init__(self, name, suffix, high_th, low_th, events_handler):
        self.name = name
        self.suffix = suffix
        self.high_th = high_th
        self.low_th = low_th
        self.events_handler = events_handler

        self.hot = False
        self.current = None

    def update(self, value):
        if value is None:
            return

        if value > self.high_th and not self.hot:
            self.events_handler.log_warning(
                "{0} is high: {1}{2}".format(self.name, value, self.suffix))
            self.hot = True

        if value < self.low_th and self.hot:
            self.hot = False

        self.current = value


class CTRexInfoGenerator(object):
    """
    This object is responsible of generating stats and information from objects maintained at
    STLClient and the ports.
    """

    def __init__(self, global_stats_ref, ports_dict_ref, rx_stats_ref, latency_stats_ref,
                 util_stats_ref, xstats_ref, async_monitor, pgid_ref, stl_client):
        self._global_stats = global_stats_ref
        self._ports_dict = ports_dict_ref
        self._rx_stats_ref = rx_stats_ref
        self._latency_stats_ref = latency_stats_ref
        self._util_stats_ref = util_stats_ref
        self._xstats_ref = xstats_ref
        self._async_monitor = async_monitor
        self.pgid_ref = pgid_ref
        self._stl_client = stl_client
        self.latency_window_size = 14
        self.max_hist = {}
        self.max_hist_index = 0

    def clear_stats(self):
        self.max_hist = {}
        self.max_hist_index = 0

    def generate_single_statistic(self, port_id_list, statistic_type):
        if statistic_type == GLOBAL_STATS:
            return self._generate_global_stats()

        elif statistic_type == PORT_STATS:
            return self._generate_port_stats(port_id_list)

        elif statistic_type == PORT_GRAPH:
            return self._generate_port_graph(port_id_list)

        elif statistic_type == PORT_STATUS:
            return self._generate_port_status(port_id_list)

        elif statistic_type == STREAMS_STATS:
            return self._generate_streams_stats()

        elif statistic_type == LATENCY_STATS:
            return self._generate_latency_stats()

        elif statistic_type == LATENCY_HISTOGRAM:
            return self._generate_latency_histogram()

        elif statistic_type == CPU_STATS:
            return self._generate_cpu_util_stats()

        elif statistic_type == MBUF_STATS:
            return self._generate_mbuf_util_stats()

        elif statistic_type == EXTENDED_STATS:
            return self._generate_xstats(port_id_list, include_zero_lines=False)

        elif statistic_type == EXTENDED_INC_ZERO_STATS:
            return self._generate_xstats(port_id_list, include_zero_lines=True)

        else:
            # ignore by returning empty object
            return {}

    def generate_streams_info(self, port_id_list, stream_id_list):
        relevant_ports = self.__get_relevant_ports(port_id_list)
        return_data = OrderedDict()

        for port_obj in relevant_ports:
            streams_data = self._generate_single_port_streams_info(
                port_obj, stream_id_list)
            if not streams_data:
                continue
            hdr_key = "Port {port}:".format(port=port_obj.port_id)

            # TODO: test for other ports with same stream structure, and join them
            return_data[hdr_key] = streams_data

        return return_data

    def _generate_global_stats(self):
        global_stats = self._global_stats

        stats_data_left = OrderedDict([("connection", "{host}, Port {port}"
                                        .format(host=global_stats.connection_info.get("server"),
                                                port=global_stats.connection_info.get("sync_port")
                                                )),
                                       ("version", "{ver}".format(
                                           ver=global_stats.server_version.get("version", "N/A"))),

                                       ("cpu_util.", "{0}% @ {2} cores({3} per port) {1}"
                                        .format(format_threshold(round_float(global_stats
                                                                             .get("m_cpu_util")),
                                                [85, 100], [0, 85]),
                                                global_stats.get_trend_gui("m_cpu_util",
                                                                           use_raw=True),
                                                global_stats.system_info.get('dp_core_count'),
                                                global_stats.system_info.get(
                                                    'dp_core_count_per_port'),)),

                                       ("rx_cpu_util.", "{0}% / {1} {2}"
                                        .format(format_threshold(round_float(global_stats
                                                .get("m_rx_cpu_util")), [85, 100], [0, 85]),
                                                global_stats.get("m_rx_core_pps",
                                                                 format=True, suffix="pkt/sec"),
                                                global_stats.get_trend_gui("m_rx_cpu_util",
                                                                           use_raw=True),)),

                                       ("async_util.", "{0}% / {1}"
                                        .format(format_threshold(round_float(self._async_monitor
                                                .get_cpu_util()), [85, 100], [0, 85]),
                                                format_num(self._async_monitor.get_bps() / 8.0,
                                                           suffix="B/sec"))),
                                       ])

        stats_data_right = OrderedDict([
            ("total_tx_L2", "{0} {1}".format(global_stats.get("m_tx_bps", format=True,
                                                              suffix="b/sec"),
                                             global_stats.get_trend_gui("m_tx_bps"))),

            ("total_tx_L1", "{0} {1}".format(global_stats.get("m_tx_bps_L1", format=True,
                                                              suffix="b/sec"),
                                             global_stats.get_trend_gui("m_tx_bps_L1"))),

            ("total_rx", "{0} {1}".format(global_stats.get("m_rx_bps", format=True, suffix="b/sec"),
                                          global_stats.get_trend_gui("m_rx_bps"))),

            ("total_pps", "{0} {1}".format(global_stats.get("m_tx_pps", format=True,
                                                            suffix="pkt/sec"),
                                           global_stats.get_trend_gui("m_tx_pps"))),

            ("drop_rate", "{0}".format(format_num(global_stats.get("m_rx_drop_bps"),
                                                  suffix='b/sec',
                                                  opts='green' if(global_stats.get("m_rx_drop_bps")
                                                                  == 0) else 'red'),
                                       )),

            ("queue_full", "{0}".format(format_num(global_stats.get_rel("m_total_queue_full"),
                                                   suffix='pkts',
                                                   compact=False,
                                                   opts='green'
                                                   if(global_stats.get_rel("m_total_queue_full")
                                                      == 0) else 'red'))),
        ])

        # build table representation
        stats_table = text_tables.TRexTextInfo()
        stats_table.set_cols_align(["l", "l"])
        stats_table.set_deco(0)
        stats_table.set_cols_width([50, 45])
        max_lines = max(len(stats_data_left), len(stats_data_right))
        for line_num in range(max_lines):
            row = []
            if line_num < len(stats_data_left):
                key = list(stats_data_left.keys())[line_num]
                row.append('{:<12} : {}'.format(key, stats_data_left[key]))
            else:
                row.append('')
            if line_num < len(stats_data_right):
                key = list(stats_data_right.keys())[line_num]
                row.append('{:<12} : {}'.format(key, stats_data_right[key]))
            else:
                row.append('')
            stats_table.add_row(row)

        return {"global_statistics": ExportableStats(None, stats_table)}

    def _generate_streams_stats(self):
        all_pg_ids = self._stl_client.get_active_pgids()
        # Display data for at most 4 pgids. If there are latency PG IDs, use them first
        pg_ids = all_pg_ids['latency'][:4]
        pg_ids += all_pg_ids['flow_stats'][:4 - len(pg_ids)]
        stream_count = len(pg_ids)
        # data = self._stl_client.get_pgid_stats(pg_ids)
        # if 'flow_stats' in data:
        #     flow_stats = data['flow_stats']

        sstats_data = OrderedDict([('Tx pps',  []),
                                   ('Tx bps L2',      []),
                                   ('Tx bps L1',      []),
                                   ('---', [''] * stream_count),
                                   ('Rx pps',      []),
                                   ('Rx bps',      []),
                                   ('----', [''] * stream_count),
                                   ('opackets',    []),
                                   ('ipackets',    []),
                                   ('obytes',      []),
                                   ('ibytes',      []),
                                   ('-----', [''] * stream_count),
                                   ('tx_pkts',     []),
                                   ('rx_pkts',     []),
                                   ('tx_bytes',    []),
                                   ('rx_bytes',    [])
                                   ])

        # maximum 4
        for pg_id in pg_ids:
            sstats_data['Tx pps'].append(self.pgid_ref.get(
                ['flow_stats', pg_id, 'tx_pps', 'total'], format=True, suffix="pps"))
            sstats_data['Tx bps L2'].append(self.pgid_ref.get(
                ['flow_stats', pg_id, 'tx_bps', 'total'], format=True, suffix="bps"))
            sstats_data['Tx bps L1'].append(self.pgid_ref.get(
                ['flow_stats', pg_id, 'tx_bps_l1', 'total'], format=True, suffix="bps"))
            sstats_data['Rx pps'].append(self.pgid_ref.get(
                ['flow_stats', pg_id, 'rx_pps', 'total'], format=True, suffix="pps"))
            sstats_data['Rx bps'].append(self.pgid_ref.get(
                ['flow_stats', pg_id, 'rx_bps', 'total'], format=True, suffix="bps"))
            sstats_data['opackets'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'tx_pkts', 'total']))
            sstats_data['ipackets'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'rx_pkts', 'total']))
            sstats_data['obytes'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'tx_bytes', 'total']))
            sstats_data['ibytes'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'rx_bytes', 'total']))
            sstats_data['tx_pkts'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'tx_pkts', 'total'], format=True, suffix="pkts"))
            sstats_data['rx_pkts'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'rx_pkts', 'total'], format=True, suffix="pkts"))
            sstats_data['tx_bytes'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'tx_bytes', 'total'], format=True, suffix="B"))
            sstats_data['rx_bytes'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'rx_bytes', 'total'], format=True, suffix="B"))

        stats_table = text_tables.TRexTextTable()
        stats_table.set_cols_align(["l"] + ["r"] * stream_count)
        stats_table.set_cols_width([10] + [17] * stream_count)
        stats_table.set_cols_dtype(['t'] + ['t'] * stream_count)

        disp_str = {'tx_pkts': "opackets ", 'rx_pkts': "ipackets ",
                    'rx_bytes': "ibytes ", 'tx_bytes': "obytes "}
        sstats_new_data = OrderedDict([(disp_str[k], v) if k in disp_str else(
            k, v) for k, v in list(sstats_data.items())])
        stats_table.add_rows([[k] + v
                              for k, v in list(sstats_new_data.items())],
                             header=False)

        header = ["PG ID"] + [key for key in pg_ids]
        stats_table.header(header)

        return {"streams_statistics": ExportableStats(sstats_data, stats_table)}

    def _generate_latency_stats(self):
        all_pg_ids = self._stl_client.get_active_pgids()
        # Display data for at most 5 pgids.
        pg_ids = all_pg_ids['latency'][:5]
        to_delete = []
        for id in list(self.max_hist.keys()):
            if id not in pg_ids:
                to_delete.append(id)

        for id in to_delete:
            del self.max_hist[id]

        for id in pg_ids:
            if id not in list(self.max_hist.keys()):
                self.max_hist[id] = [-1] * self.latency_window_size

        stream_count = len(pg_ids)
        # data = self._stl_client.get_pgid_stats(pg_ids)
        # if 'latency' in data:
        #     lat_stats = data['latency']
        # if 'flow_stats' in data:
        #    flow_stats = data['flow_stats']

        lstats_data = OrderedDict([('TX pkts',       []),
                                   ('RX pkts',       []),
                                   ('Max latency',   []),
                                   ('Avg latency',   []),
                                   ('-- Window --', [''] * stream_count),
                                   ('Last max',     []),
                                   ] + [('Last-%s' % i, [])
                                        for i in range(1, self.latency_window_size)] + [
                                  ('---', [''] * stream_count),
                                  ('Jitter',        []),
                                  ('----', [''] * stream_count),
                                  ('Errors',        []),
                                  ])

        for pg_id in pg_ids:
            last_max = self.pgid_ref.get(
                ['latency', pg_id, 'latency', 'last_max'])
            lstats_data['TX pkts'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'tx_pkts', 'total']))
            lstats_data['RX pkts'].append(self.pgid_ref.get_rel(
                ['flow_stats', pg_id, 'rx_pkts', 'total']))
            lstats_data['Avg latency'].append(
                try_int(self.pgid_ref.get(['latency', pg_id, 'latency', 'average'])))
            lstats_data['Max latency'].append(
                try_int(self.pgid_ref.get(['latency', pg_id, 'latency', 'total_max'])))
            lstats_data['Last max'].append(last_max)
            self.max_hist[pg_id][self.max_hist_index] = last_max
            for i in range(1, self.latency_window_size):
                val = self.max_hist[pg_id][(
                    self.max_hist_index - i) % self.latency_window_size]
                if val != -1:
                    lstats_data['Last-%s' % i].append(val)
                else:
                    lstats_data['Last-%s' % i].append(" ")
            lstats_data['Jitter'].append(self.pgid_ref.get(
                ['latency', pg_id, 'latency', 'jitter']))
            errors = 0
            seq_too_low = self.pgid_ref.get_rel(
                ['latency', pg_id, 'err_cntrs', 'seq_too_low'])
            if is_integer(seq_too_low):
                errors += seq_too_low
            seq_too_high = self.pgid_ref.get_rel(
                ['latency', pg_id, 'err_cntrs', 'seq_too_high'])
            if is_integer(seq_too_high):
                errors += seq_too_high
            lstats_data['Errors'].append(format_num(errors,
                                                    opts='green' if errors == 0 else 'red'))
        self.max_hist_index += 1
        self.max_hist_index %= self.latency_window_size

        stats_table = text_tables.TRexTextTable()
        stats_table.set_cols_align(["l"] + ["r"] * stream_count)
        stats_table.set_cols_width([12] + [14] * stream_count)
        stats_table.set_cols_dtype(['t'] + ['t'] * stream_count)
        stats_table.add_rows([[k] + v
                              for k, v in list(lstats_data.items())],
                             header=False)

        header = ["PG ID"] + [key for key in pg_ids]
        stats_table.header(header)

        return {"latency_statistics": ExportableStats(lstats_data, stats_table)}

    def _generate_latency_histogram(self):
        all_pg_ids = self._stl_client.get_active_pgids()
        # Display data for at most 5 pgids.
        pg_ids = all_pg_ids['latency'][:5]
        data = self._stl_client.get_pgid_stats(pg_ids)
        if 'latency' in data:
            lat_stats = data['latency']

        max_histogram_size = 17
        merged_histogram = {}
        for pg_id in pg_ids:
            merged_histogram.update(lat_stats[pg_id]['latency']['histogram'])
        histogram_size = min(max_histogram_size, len(merged_histogram))

        stream_count = len(pg_ids)
        stats_table = text_tables.TRexTextTable()
        stats_table.set_cols_align(["l"] + ["r"] * stream_count)
        stats_table.set_cols_width([12] + [14] * stream_count)
        stats_table.set_cols_dtype(['t'] + ['t'] * stream_count)

        for i in range(max_histogram_size - histogram_size):
            if i == 0 and not merged_histogram:
                stats_table.add_row(['  No Data'] + [' '] * stream_count)
            else:
                stats_table.add_row([' '] * (stream_count + 1))
        for key in list(reversed(sorted(merged_histogram.keys())))[:histogram_size]:
            hist_vals = []
            for pg_id in pg_ids:
                hist_vals.append(
                    lat_stats[pg_id]['latency']['histogram'].get(key, ' '))
            stats_table.add_row([key] + hist_vals)

        stats_table.add_row(['- Counters -'] + [' '] * stream_count)
        err_cntrs_dict = OrderedDict()
        for pg_id in pg_ids:
            for err_cntr in sorted(lat_stats[pg_id]['err_cntrs'].keys()):
                if err_cntr not in err_cntrs_dict:
                    err_cntrs_dict[err_cntr] = [
                        lat_stats[pg_id]['err_cntrs'][err_cntr]]
                else:
                    err_cntrs_dict[err_cntr].append(
                        lat_stats[pg_id]['err_cntrs'][err_cntr])
        for err_cntr, val_list in list(err_cntrs_dict.items()):
            stats_table.add_row([err_cntr] + val_list)
        header = ["PG ID"] + [key for key in pg_ids]
        stats_table.header(header)
        return {"latency_histogram": ExportableStats(None, stats_table)}

    def _generate_cpu_util_stats(self):
        util_stats = self._util_stats_ref.get_stats(use_1sec_cache=True)

        stats_table = text_tables.TRexTextTable()
        if util_stats:
            if 'cpu' not in util_stats:
                raise Exception(
                    "Excepting 'cpu' section in stats %s" % util_stats)
            cpu_stats = util_stats['cpu']
            hist_len = len(cpu_stats[0]["history"])
            avg_len = min(5, hist_len)
            show_len = min(15, hist_len)
            stats_table.header(['Thread', 'Avg', 'Latest'] +
                               list(range(-1, 0 - show_len, -1)))
            stats_table.set_cols_align(['l'] + ['r'] * (show_len + 1))
            stats_table.set_cols_width([10, 3, 6] + [3] * (show_len - 1))
            stats_table.set_cols_dtype(['t'] * (show_len + 2))

            for i in range(min(18, len(cpu_stats))):
                history = cpu_stats[i]["history"]
                ports = cpu_stats[i]["ports"]
                avg = int(round(sum(history[:avg_len]) / avg_len))

                # decode active ports for core
                if ports == [-1, -1]:
                    interfaces = "(IDLE)"
                elif -1 not in ports:
                    interfaces = "({:},{:})".format(ports[0], ports[1])
                else:
                    interfaces = "({:})".format(
                        ports[0] if ports[0] != -1 else ports[1])

                thread = "{:2} {:^7}".format(i, interfaces)
                stats_table.add_row([thread, avg] + history[:show_len])
        else:
            stats_table.add_row(['No Data.'])
        return {'cpu_util(%)': ExportableStats(None, stats_table)}

    def _generate_mbuf_util_stats(self):
        util_stats = self._util_stats_ref.get_stats(use_1sec_cache=True)
        stats_table = text_tables.TRexTextTable()
        if util_stats:
            if 'mbuf_stats' not in util_stats:
                raise Exception(
                    "Excepting 'mbuf_stats' section in stats %s" % util_stats)
            mbuf_stats = util_stats['mbuf_stats']
            for mbufs_per_socket in list(mbuf_stats.values()):
                first_socket_mbufs = mbufs_per_socket
                break
            if not self._util_stats_ref.mbuf_types_list:
                mbuf_keys = list(first_socket_mbufs.keys())
                mbuf_keys.sort(key=get_number_of_bytes)
                self._util_stats_ref.mbuf_types_list = mbuf_keys
            types_len = len(self._util_stats_ref.mbuf_types_list)
            stats_table.set_cols_align(['l'] + ['r'] * (types_len + 1))
            stats_table.set_cols_width([10] + [7] * (types_len + 1))
            stats_table.set_cols_dtype(['t'] * (types_len + 2))
            stats_table.header(
                [''] + self._util_stats_ref.mbuf_types_list + ['RAM(MB)'])
            total_list = []
            sum_totals = 0
            for mbuf_type in self._util_stats_ref.mbuf_types_list:
                sum_totals += first_socket_mbufs[mbuf_type][1] * \
                    get_number_of_bytes(mbuf_type) + 64
                total_list.append(first_socket_mbufs[mbuf_type][1])
            sum_totals *= len(list(mbuf_stats.values()))
            total_list.append(int(sum_totals/1e6))
            stats_table.add_row(['Total:'] + total_list)
            stats_table.add_row(['Used:'] + [''] * (types_len + 1))
            for socket_name in sorted(list(mbuf_stats.keys())):
                mbufs = mbuf_stats[socket_name]
                socket_show_name = socket_name.replace(
                    'cpu-', '').replace('-', ' ').capitalize() + ':'
                sum_used = 0
                used_list = []
                percentage_list = []
                for mbuf_type in self._util_stats_ref.mbuf_types_list:
                    used = mbufs[mbuf_type][1] - mbufs[mbuf_type][0]
                    sum_used += used * get_number_of_bytes(mbuf_type) + 64
                    used_list.append(used)
                    percentage_list.append('%s%%' % int(
                        100 * used / mbufs[mbuf_type][1]))
                used_list.append(int(sum_used/1e6))
                stats_table.add_row([socket_show_name] + used_list)
                stats_table.add_row(['Percent:'] + percentage_list + [''])
        else:
            stats_table.add_row(['No Data.'])
        return {'mbuf_util': ExportableStats(None, stats_table)}

    def _generate_xstats(self, port_id_list, include_zero_lines=False):
        relevant_ports = [
            port.port_id for port in self.__get_relevant_ports(port_id_list)]
        # get the data on relevant ports
        xstats_data = OrderedDict()
        for port_id in relevant_ports:
            for key, val in list(self._xstats_ref.get_stats(port_id).items()):
                if key not in xstats_data:
                    xstats_data[key] = []
                xstats_data[key].append(val)

        # put into table
        stats_table = text_tables.TRexTextTable()
        stats_table.header(['Name:'] + ['Port %s:' %
                                        port_id for port_id in relevant_ports])
        stats_table.set_cols_align(['l'] + ['r'] * len(relevant_ports))
        stats_table.set_cols_width([30] + [15] * len(relevant_ports))
        stats_table.set_cols_dtype(['t'] * (len(relevant_ports) + 1))
        for key, arr in list(xstats_data.items()):
            if include_zero_lines or list([_f for _f in arr if _f]):
                key = key[:28]
                stats_table.add_row([key] + arr)
        return {'xstats:': ExportableStats(None, stats_table)}

    @staticmethod
    def _get_rational_block_char(value, range_start, interval):
        # in Konsole, utf-8 is sometimes printed with artifacts, return ascii for now
        # return 'X' if value >= range_start + float(interval) / 2 else ' '

        if sys.__stdout__.encoding != 'UTF-8':
            return 'X' if value >= range_start + float(interval) / 2 else ' '

        value -= range_start
        ratio = float(value) / interval
        if ratio <= 0.0625:
            return ' '         # empty block
        if ratio <= 0.1875:
            return '\\u2581'    # 1/8
        if ratio <= 0.3125:
            return '\\u2582'    # 2/8
        if ratio <= 0.4375:
            return '\\u2583'    # 3/8
        if ratio <= 0.5625:
            return '\\u2584'    # 4/8
        if ratio <= 0.6875:
            return '\\u2585'    # 5/8
        if ratio <= 0.8125:
            return '\\u2586'    # 6/8
        if ratio <= 0.9375:
            return '\\u2587'    # 7/8
        return '\\u2588'        # full block

    def _generate_port_graph(self, port_id_list):
        relevant_port = self.__get_relevant_ports(port_id_list)[0]
        hist_len = len(relevant_port.port_stats.history)
        hist_maxlen = relevant_port.port_stats.history.maxlen
        util_tx_hist = [0] * (hist_maxlen - hist_len) + [round(
            relevant_port.port_stats.history[i]['tx_percentage']) for i in range(hist_len)]
        util_rx_hist = [0] * (hist_maxlen - hist_len) + [round(
            relevant_port.port_stats.history[i]['rx_percentage']) for i in range(hist_len)]

        stats_table = text_tables.TRexTextTable()
        stats_table.header([' Util(%)', 'TX', 'RX'])
        stats_table.set_cols_align(['c', 'c', 'c'])
        stats_table.set_cols_width([8, hist_maxlen, hist_maxlen])
        stats_table.set_cols_dtype(['t', 't', 't'])

        for y in range(95, -1, -5):
            stats_table.add_row([y, ''.join([self._get_rational_block_char(util_tx, y, 5)
                                for util_tx in util_tx_hist]),
                                    ''.join([self._get_rational_block_char(util_rx, y, 5)
                                            for util_rx in util_rx_hist])])

        return {"port_graph": ExportableStats({}, stats_table)}

    def _generate_port_stats(self, port_id_list):
        relevant_ports = self.__get_relevant_ports(port_id_list)

        return_stats_data = {}
        per_field_stats = OrderedDict([("owner", []),
                                       ('link', []),
                                       ("state", []),
                                       ("speed", []),
                                       ("CPU util.", []),
                                       ("--", []),
                                       ("Tx bps L2", []),
                                       ("Tx bps L1", []),
                                       ("Tx pps", []),
                                       ("Line Util.", []),

                                       ("---", []),
                                       ("Rx bps", []),
                                       ("Rx pps", []),

                                       ("----", []),
                                       ("opackets", []),
                                       ("ipackets", []),
                                       ("obytes", []),
                                       ("ibytes", []),
                                       ("tx-pkts", []),
                                       ("rx-pkts", []),
                                       ("tx-bytes", []),
                                       ("rx-bytes", []),

                                       ("-----", []),
                                       ("oerrors", []),
                                       ("ierrors", []),

                                       ])

        total_stats = CPortStats(None)

        for port_obj in relevant_ports:
            # fetch port data
            port_stats = port_obj.generate_port_stats()

            total_stats += port_obj.port_stats

            # populate to data structures
            return_stats_data[port_obj.port_id] = port_stats
            self.__update_per_field_dict(port_stats, per_field_stats)

        total_cols = len(relevant_ports)
        header = ["port"] + [port.port_id for port in relevant_ports]

        if(total_cols > 1):
            self.__update_per_field_dict(
                total_stats.generate_stats(), per_field_stats)
            header += ['total']
            total_cols += 1

        stats_table = text_tables.TRexTextTable()
        stats_table.set_cols_align(["l"] + ["r"] * total_cols)
        stats_table.set_cols_width([10] + [17] * total_cols)
        stats_table.set_cols_dtype(['t'] + ['t'] * total_cols)

        disp_str = {'tx-pkts': "opackets ", 'rx-pkts': "ipackets ",
                    'rx-bytes': "ibytes ", 'tx-bytes': "obytes "}
        disp_data = OrderedDict([(disp_str[k], v) if k in disp_str else(
            k, v) for k, v in list(per_field_stats.items())])

        stats_table.add_rows([[k] + v
                              for k, v in list(disp_data.items())],
                             header=False)

        stats_table.header(header)

        return {"port_statistics": ExportableStats(return_stats_data, stats_table)}

    def _generate_port_status(self, port_id_list):
        relevant_ports = self.__get_relevant_ports(port_id_list)

        return_stats_data = {}
        per_field_status = OrderedDict([("driver", []),
                                        ("description", []),
                                        ("link status", []),
                                        ("link speed", []),
                                        ("port status", []),
                                        ("promiscuous", []),
                                        ("multicast", []),
                                        ("flow ctrl", []),
                                        ("--", []),
                                        ("layer mode", []),
                                        ("src IPv4", []),
                                        ("src MAC", []),
                                        ("---", []),
                                        ("Destination", []),
                                        ("ARP Resolution", []),
                                        ("----", []),
                                        ("VLAN", []),
                                        ("-----", []),
                                        ("PCI Address", []),
                                        ("NUMA Node", []),
                                        ("-----", []),
                                        ("RX Filter Mode", []),
                                        ("RX Queueing", []),
                                        ("Grat ARP", []),
                                        ]
                                       )

        for port_obj in relevant_ports:
            # fetch port data
            # port_stats = self._async_stats.get_port_stats(port_obj.port_id)
            port_status = port_obj.generate_port_status()

            # populate to data structures
            return_stats_data[port_obj.port_id] = port_status

            self.__update_per_field_dict(port_status, per_field_status)

        stats_table = text_tables.TRexTextTable()
        stats_table.set_cols_align(["l"] + ["c"]*len(relevant_ports))
        stats_table.set_cols_width([15] + [20] * len(relevant_ports))

        stats_table.add_rows([[k] + v
                              for k, v in list(per_field_status.items())],
                             header=False)
        stats_table.header(["port"] + [port.port_id
                                       for port in relevant_ports])

        return {"port_status": ExportableStats(return_stats_data, stats_table)}

    def _generate_single_port_streams_info(self, port_obj, stream_id_list):

        return_streams_data = port_obj.generate_loaded_streams_sum()

        if not return_streams_data.get("streams"):
            # we got no streams available
            return None

        # FORMAT VALUES ON DEMAND

        # because we mutate this - deep copy before
        return_streams_data = copy.deepcopy(return_streams_data)

        p_type_field_len = 0

        for stream_id, stream_id_sum in list(return_streams_data['streams'].items()):
            stream_id_sum['packet_type'] = self._trim_packet_headers(
                stream_id_sum['packet_type'], 30)
            p_type_field_len = max(p_type_field_len, len(
                stream_id_sum['packet_type']))

        info_table = text_tables.TRexTextTable()
        info_table.set_cols_align(
            ["c"] + ["l"] + ["r"] + ["c"] + ["r"] + ["c"])
        info_table.set_cols_width(
            [10] + [p_type_field_len] + [8] + [16] + [15] + [12])
        info_table.set_cols_dtype(
            ["t"] + ["t"] + ["t"] + ["t"] + ["t"] + ["t"])

        info_table.add_rows([list(v.values())
                             for k, v in list(return_streams_data['streams'].items())],
                            header=False)
        info_table.header(["ID", "packet type", "length",
                           "mode", "rate", "next stream"])

        return ExportableStats(return_streams_data, info_table)

    def __get_relevant_ports(self, port_id_list):
        # fetch owned ports
        ports = [port_obj
                 for _, port_obj in list(self._ports_dict.items())
                 if port_obj.port_id in port_id_list]

        # display only the first FOUR options, by design
        if len(ports) > 4:
            # self.logger is not defined
            # self.logger.log(format_text("[WARNING]: ", 'magenta', 'bold'),
            #                 format_text("displaying up to 4 ports", 'magenta'))
            ports = ports[:4]
        return ports

    def __update_per_field_dict(self, dict_src_data, dict_dest_ref):
        for key, val in list(dict_src_data.items()):
            if key in dict_dest_ref:
                dict_dest_ref[key].append(val)

    @staticmethod
    def _trim_packet_headers(headers_str, trim_limit):
        if len(headers_str) < trim_limit:
            # do nothing
            return headers_str
        else:
            return(headers_str[:trim_limit-3] + "...")


class CTRexStats(object):
    """ This is an abstract class to represent a stats object """

    def __init__(self):
        self.reference_stats = {}
        self.latest_stats = {}
        self.last_update_ts = time.time()
        self.history = deque(maxlen=47)
        self.lock = threading.Lock()
        self.has_baseline = False

    # ####### abstract methods ##########

    # get stats for user / API
    def get_stats(self):
        raise NotImplementedError()

    # generate format stats(for TUI)
    def generate_stats(self):
        raise NotImplementedError()

    # called when a snapshot arrives - add more fields
    def _update(self, snapshot, baseline):
        raise NotImplementedError()

    # ####### END abstract methods ##########

    def update(self, snapshot, baseline):

        # no update is valid before baseline
        if not self.has_baseline and not baseline:
            return

        # call the underlying method
        rc = self._update(snapshot)
        if not rc:
            return

        # sync one time
        if not self.has_baseline and baseline:
            self.reference_stats = copy.deepcopy(self.latest_stats)
            self.has_baseline = True

        # save history
        with self.lock:
            self.history.append(self.latest_stats)

    def clear_stats(self):
        self.reference_stats = copy.deepcopy(self.latest_stats)
        self.history.clear()

    def invalidate(self):
        self.latest_stats = {}

    def _get(self, src, field, default=None):
        if isinstance(field, list):
            # deep
            value = src
            for level in field:
                if level not in value:
                    return default
                value = value[level]
        else:
            # flat
            if field not in src:
                return default
            value = src[field]

        return value

    def get(self, field, format=False, suffix="", opts=None):
        value = self._get(self.latest_stats, field)
        if value is None:
            return 'N/A'

        return value if not format else format_num(value, suffix=suffix, opts=opts)

    def get_rel(self, field, format=False, suffix=""):
        ref_value = self._get(self.reference_stats, field)
        latest_value = self._get(self.latest_stats, field)

        # latest value is an aggregation - must contain the value
        if latest_value is None:
            return 'N/A'

        if ref_value is None:
            ref_value = 0

        value = latest_value - ref_value

        return value if not format else format_num(value, suffix)

    # get trend for a field

    def get_trend(self, field, use_raw=False, percision=10.0):
        if field not in self.latest_stats:
            return 0

        # not enough history - no trend
        if len(self.history) < 5:
            return 0

        # absolute value is too low 0 considered noise
        if self.latest_stats[field] < percision:
            return 0

        # must lock, deque is not thread-safe for iteration
        with self.lock:
            field_samples = [sample[field]
                             for sample in list(self.history)[-5:]]

        if use_raw:
            return calculate_diff_raw(field_samples)
        else:
            return calculate_diff(field_samples)

    def get_trend_gui(self, field, show_value=False, use_raw=False, up_color='red',
                      down_color='green'):
        v = self.get_trend(field, use_raw)

        value = abs(v)

        # use arrows if utf-8 is supported
        if sys.__stdout__.encoding == 'UTF-8':
            arrow = '\\u25b2' if v > 0 else '\\u25bc'
        else:
            arrow = ''

        if sys.version_info < (3, 0):
            arrow = arrow.encode('utf-8')

        color = up_color if v > 0 else down_color

        # change in 1% is not meaningful
        if value < 1:
            return ""

        elif value > 5:

            if show_value:
                return format_text("{0}{0}{0} {1:.2f}%".format(arrow, v), color)
            else:
                return format_text("{0}{0}{0}".format(arrow), color)

        elif value > 2:

            if show_value:
                return format_text("{0}{0} {1:.2f}%".format(arrow, v), color)
            else:
                return format_text("{0}{0}".format(arrow), color)

        else:
            if show_value:
                return format_text("{0} {1:.2f}%".format(arrow, v), color)
            else:
                return format_text("{0}".format(arrow), color)


class CGlobalStats(CTRexStats):

    def __init__(self, connection_info, server_version, ports_dict_ref, events_handler):
        super(CGlobalStats, self).__init__()

        self.connection_info = connection_info
        self.server_version = server_version
        self._ports_dict = ports_dict_ref
        self.events_handler = events_handler

        self.watched_cpu_util = WatchedField(
            'CPU util.', '%', 85, 60, events_handler)
        self.watched_rx_cpu_util = WatchedField(
            'RX core util.', '%', 85, 60, events_handler)

    def get_stats(self):
        stats = {}

        # absolute
        stats['cpu_util'] = self.get("m_cpu_util")
        stats['rx_cpu_util'] = self.get("m_rx_cpu_util")
        stats['bw_per_core'] = self.get("m_bw_per_core")

        stats['tx_bps'] = self.get("m_tx_bps")
        stats['tx_pps'] = self.get("m_tx_pps")

        stats['rx_bps'] = self.get("m_rx_bps")
        stats['rx_pps'] = self.get("m_rx_pps")
        stats['rx_drop_bps'] = self.get("m_rx_drop_bps")

        # relatives
        stats['queue_full'] = self.get_rel("m_total_queue_full")

        return stats

    def _update(self, snapshot):
        # L1 bps
        bps = snapshot.get("m_tx_bps")
        pps = snapshot.get("m_tx_pps")

        snapshot['m_tx_bps_L1'] = calc_bps_L1(bps, pps)

        # simple...
        self.latest_stats = snapshot

        self.watched_cpu_util.update(snapshot.get('m_cpu_util'))
        self.watched_rx_cpu_util.update(snapshot.get('m_rx_cpu_util'))

        return True


class CPortStats(CTRexStats):

    def __init__(self, port_obj):
        super(CPortStats, self).__init__()
        self._port_obj = port_obj

    @staticmethod
    def __merge_dicts(target, src):
        for k, v in list(src.items()):
            if k in target:
                target[k] += v
            else:
                target[k] = v

    def __add__(self, x):
        if not isinstance(x, CPortStats):
            raise TypeError("cannot add non stats object to stats")

        # main stats
        if not self.latest_stats:
            self.latest_stats = {}

        self.__merge_dicts(self.latest_stats, x.latest_stats)

        # reference stats
        if x.reference_stats:
            if not self.reference_stats:
                self.reference_stats = x.reference_stats.copy()
            else:
                self.__merge_dicts(self.reference_stats, x.reference_stats)

        # history - should be traverse with a lock
        with self.lock, x.lock:
            if not self.history:
                self.history = copy.deepcopy(x.history)
            else:
                for h1, h2 in zip(self.history, x.history):
                    self.__merge_dicts(h1, h2)

        return self

    # for port we need to do something smarter
    def get_stats(self):
        stats = {}

        stats['opackets'] = self.get_rel("opackets")
        stats['ipackets'] = self.get_rel("ipackets")
        stats['obytes'] = self.get_rel("obytes")
        stats['ibytes'] = self.get_rel("ibytes")
        stats['oerrors'] = self.get_rel("oerrors")
        stats['ierrors'] = self.get_rel("ierrors")

        stats['tx_bps'] = self.get("m_total_tx_bps")
        stats['tx_pps'] = self.get("m_total_tx_pps")
        stats['tx_bps_L1'] = self.get("m_total_tx_bps_L1")
        stats['tx_util'] = self.get("m_tx_util")

        stats['rx_bps'] = self.get("m_total_rx_bps")
        stats['rx_pps'] = self.get("m_total_rx_pps")
        stats['rx_bps_L1'] = self.get("m_total_rx_bps_L1")
        stats['rx_util'] = self.get("m_rx_util")

        return stats

    def _update(self, snapshot):
        speed = self._port_obj.get_speed_bps()

        # L1 bps
        tx_bps = snapshot.get("m_total_tx_bps")
        tx_pps = snapshot.get("m_total_tx_pps")
        rx_bps = snapshot.get("m_total_rx_bps")
        rx_pps = snapshot.get("m_total_rx_pps")
        ts_diff = 0.5  # TODO: change this to real ts diff from server

        bps_tx_L1 = calc_bps_L1(tx_bps, tx_pps)
        bps_rx_L1 = calc_bps_L1(rx_bps, rx_pps)

        snapshot['m_total_tx_bps_L1'] = bps_tx_L1
        if speed:
            snapshot['m_tx_util'] = (bps_tx_L1 / speed) * 100.0
        else:
            snapshot['m_tx_util'] = 0

        snapshot['m_total_rx_bps_L1'] = bps_rx_L1
        if speed:
            snapshot['m_rx_util'] = (bps_rx_L1 / speed) * 100.0
        else:
            snapshot['m_rx_util'] = 0

        # TX line util not smoothed
        diff_tx_pkts = snapshot.get(
            'opackets', 0) - self.latest_stats.get('opackets', 0)
        diff_tx_bytes = snapshot.get(
            'obytes', 0) - self.latest_stats.get('obytes', 0)
        tx_bps_L1 = calc_bps_L1(8.0 * diff_tx_bytes /
                                ts_diff, float(diff_tx_pkts) / ts_diff)
        if speed:
            snapshot['tx_percentage'] = 100.0 * tx_bps_L1 / speed
        else:
            snapshot['tx_percentage'] = 0

        # RX line util not smoothed
        diff_rx_pkts = snapshot.get(
            'ipackets', 0) - self.latest_stats.get('ipackets', 0)
        diff_rx_bytes = snapshot.get(
            'ibytes', 0) - self.latest_stats.get('ibytes', 0)
        rx_bps_L1 = calc_bps_L1(8.0 * diff_rx_bytes /
                                ts_diff, float(diff_rx_pkts) / ts_diff)
        if speed:
            snapshot['rx_percentage'] = 100.0 * rx_bps_L1 / speed
        else:
            snapshot['rx_percentage'] = 0

        # simple...
        self.latest_stats = snapshot

        return True

    def generate_stats(self):

        port_state = self._port_obj.get_port_state_name() if self._port_obj else ""
        if port_state == "TRANSMITTING":
            port_state = format_text(port_state, 'green', 'bold')
        elif port_state == "PAUSE":
            port_state = format_text(port_state, 'magenta', 'bold')
        else:
            port_state = format_text(port_state, 'bold')

        if self._port_obj:
            link_state = 'UP' if self._port_obj.is_up() else format_text('DOWN', 'red', 'bold')
        else:
            link_state = ''

        # default rate format modifiers
        rate_format = {'bpsl1': None, 'bps': None,
                       'pps': None, 'percentage': 'bold'}

        # mark owned ports by color
        if self._port_obj:
            owner = self._port_obj.get_owner()
            rate_format[self._port_obj.last_factor_type] = ('blue', 'bold')
            if self._port_obj.is_acquired():
                owner = format_text(owner, 'green')

        else:
            owner = ''

        return {"owner": owner,
                "state": "{0}".format(port_state),
                'link': link_state,
                "speed": "%g Gb/s" % self._port_obj.get_speed_gbps() if self._port_obj else '',
                "CPU util.": "{0} {1}%".format(self.get_trend_gui("m_cpu_util", use_raw=True),
                                               format_threshold(round_float(self.get("m_cpu_util")),
                                               [85, 100], [0, 85])) if self._port_obj else '',
                "--": " ",
                "---": " ",
                "----": " ",
                "-----": " ",

                "Tx bps L1": "{0} {1}".format(self.get_trend_gui("m_total_tx_bps_L1",
                                              show_value=False),
                                              self.get("m_total_tx_bps_L1", format=True,
                                              suffix="bps", opts=rate_format['bpsl1'])),

                "Tx bps L2": "{0} {1}".format(self.get_trend_gui("m_total_tx_bps",
                                              show_value=False),
                                              self.get("m_total_tx_bps", format=True,
                                              suffix="bps", opts=rate_format['bps'])),

                "Line Util.": "{0} {1}".format(self.get_trend_gui("m_tx_util", show_value=False)
                                               if self._port_obj else "",
                                               self.get("m_tx_util", format=True, suffix="%",
                                               opts=rate_format['percentage'])
                                               if self._port_obj else ""),

                "Rx bps": "{0} {1}".format(self.get_trend_gui("m_total_rx_bps", show_value=False),
                                           self.get("m_total_rx_bps", format=True, suffix="bps")),

                "Tx pps": "{0} {1}".format(self.get_trend_gui("m_total_tx_pps", show_value=False),
                                           self.get("m_total_tx_pps", format=True, suffix="pps",
                                           opts=rate_format['pps'])),

                "Rx pps": "{0} {1}".format(self.get_trend_gui("m_total_rx_pps", show_value=False),
                                           self.get("m_total_rx_pps", format=True, suffix="pps")),

                "opackets": self.get_rel("opackets"),
                "ipackets": self.get_rel("ipackets"),
                "obytes": self.get_rel("obytes"),
                "ibytes": self.get_rel("ibytes"),

                "tx-bytes": self.get_rel("obytes", format=True, suffix="B"),
                "rx-bytes": self.get_rel("ibytes", format=True, suffix="B"),
                "tx-pkts": self.get_rel("opackets", format=True, suffix="pkts"),
                "rx-pkts": self.get_rel("ipackets", format=True, suffix="pkts"),

                "oerrors": format_num(self.get_rel("oerrors"),
                                      compact=False,
                                      opts='green' if(self.get_rel("oerrors") == 0) else 'red'),

                "ierrors": format_num(self.get_rel("ierrors"),
                                      compact=False,
                                      opts='green' if(self.get_rel("ierrors") == 0) else 'red'),

                }


class CLatencyStats(CTRexStats):
    def __init__(self, ports):
        super(CLatencyStats, self).__init__()

    # for API

    def get_stats(self):
        return copy.deepcopy(self.latest_stats)

    def _update(self, snapshot):
        if snapshot is None:
            snapshot = {}
        output = {}

        output['global'] = {}
        for field in ['bad_hdr', 'old_flow']:
            if 'global' in snapshot and field in snapshot['global']:
                output['global'][field] = snapshot['global'][field]
            else:
                output['global'][field] = 0

        # we care only about the current active keys
        pg_ids = list(filter(is_intable, list(snapshot.keys())))

        for pg_id in pg_ids:
            current_pg = snapshot.get(pg_id)
            int_pg_id = int(pg_id)
            output[int_pg_id] = {}
            output[int_pg_id]['err_cntrs'] = current_pg['err_cntrs']
            output[int_pg_id]['latency'] = {}

            if 'latency' in current_pg:
                for field in ['jitter', 'average', 'total_max', 'last_max']:
                    if field in current_pg['latency']:
                        output[int_pg_id]['latency'][field] = current_pg['latency'][field]
                    else:
                        output[int_pg_id]['latency'][field] = StatNotAvailable(
                            field)

                if 'histogram' in current_pg['latency']:
                    output[int_pg_id]['latency']['histogram'] = {int(elem): current_pg['latency']
                                                                 ['histogram'][elem]
                                                                 for elem in current_pg['latency']
                                                                 ['histogram']}
                    min_val = min(output[int_pg_id]
                                  ['latency']['histogram'].keys())
                    if min_val == 0:
                        min_val = 2
                    output[int_pg_id]['latency']['total_min'] = min_val
                else:
                    output[int_pg_id]['latency']['total_min'] = StatNotAvailable(
                        'total_min')
                    output[int_pg_id]['latency']['histogram'] = {}

        self.latest_stats = output
        return True


# RX stats objects - COMPLEX :-(
class CRxStats(CTRexStats):
    def __init__(self, ports):
        super(CRxStats, self).__init__()
        self.ports = ports

    # calculates a diff between previous snapshot
    # and current one

    def calculate_diff_sec(self, current, prev):
        if 'ts' not in current:
            raise ValueError(
                "INTERNAL ERROR: RX stats snapshot MUST contain 'ts' field")

        if prev:
            prev_ts = prev['ts']
            now_ts = current['ts']
            diff_sec = (now_ts['value'] - prev_ts['value']
                        ) / float(now_ts['freq'])
        else:
            diff_sec = 0.0

        return diff_sec

    # this is the heart of the complex

    def process_single_pg(self, current_pg, prev_pg):

        # start with the previous PG
        output = copy.deepcopy(prev_pg)

        for field in ['tx_pkts', 'tx_bytes', 'rx_pkts', 'rx_bytes']:
            # is in the first time ?(nothing in prev)
            if field not in output:
                output[field] = {}

            # does the current snapshot has this field ?
            if field in current_pg:
                for port, pv in list(current_pg[field].items()):
                    if not is_intable(port):
                        continue

                    output[field][port] = pv

            # sum up
            total = None
            for port, pv in list(output[field].items()):
                if not is_intable(port):
                    continue
                if total is None:
                    total = 0
                total += pv

            output[field]['total'] = total

        return output

    def process_snapshot(self, current, prev):

        # final output
        output = {}

        # copy timestamp field
        output['ts'] = current['ts']

        # global(not per pg_id) error counters
        output['global'] = {}
        for field in ['rx_err', 'tx_err']:
            output['global'][field] = {}
            if 'global' in current and field in current['global']:
                for port in current['global'][field]:
                    output['global'][field][int(
                        port)] = current['global'][field][port]

        # we care only about the current active keys
        pg_ids = list(filter(is_intable, list(current.keys())))

        for pg_id in pg_ids:

            current_pg = current.get(pg_id, {})

            # first time - we do not care
            if current_pg.get('first_time'):
                # new value - ignore history
                output[pg_id] = self.process_single_pg(current_pg, {})
                self.reference_stats[pg_id] = {}

                # 'dry' B/W
                self.calculate_bw_for_pg(output[pg_id])

            else:
                # aggregate the two values
                prev_pg = prev.get(pg_id, {})
                output[pg_id] = self.process_single_pg(current_pg, prev_pg)

                # calculate B/W
                diff_sec = self.calculate_diff_sec(current, prev)
                self.calculate_bw_for_pg(output[pg_id], prev_pg, diff_sec)

        # cleanp old reference values - they are dead
        ref_pg_ids = list(
            filter(is_intable, list(self.reference_stats.keys())))

        deleted_pg_ids = set(ref_pg_ids).difference(pg_ids)
        for d_pg_id in deleted_pg_ids:
            del self.reference_stats[d_pg_id]

        return output

    def calculate_bw_for_pg(self, pg_current, pg_prev=None, diff_sec=0.0):
        # no previous values
        if(not pg_prev) or not(diff_sec > 0):
            pg_current['tx_pps'] = {}
            pg_current['tx_bps'] = {}
            pg_current['tx_bps_L1'] = {}
            pg_current['tx_line_util'] = {}
            pg_current['rx_pps'] = {}
            pg_current['rx_bps'] = {}
            pg_current['rx_bps_L1'] = {}
            pg_current['rx_line_util'] = {}

            pg_current['tx_pps_lpf'] = {}
            pg_current['tx_bps_lpf'] = {}
            pg_current['tx_bps_L1_lpf'] = {}
            pg_current['rx_pps_lpf'] = {}
            pg_current['rx_bps_lpf'] = {}
            pg_current['rx_bps_L1_lpf'] = {}
            return

        # TX
        for port in list(pg_current['tx_pkts'].keys()):

            prev_tx_pps = pg_prev['tx_pps'].get(port)
            now_tx_pkts = pg_current['tx_pkts'].get(port)
            prev_tx_pkts = pg_prev['tx_pkts'].get(port)
            pg_current['tx_pps'][port], pg_current['tx_pps_lpf'][port] = self.calc_pps(
                prev_tx_pps, now_tx_pkts, prev_tx_pkts, diff_sec)

            prev_tx_bps = pg_prev['tx_bps'].get(port)
            now_tx_bytes = pg_current['tx_bytes'].get(port)
            prev_tx_bytes = pg_prev['tx_bytes'].get(port)

            pg_current['tx_bps'][port], pg_current['tx_bps_lpf'][port] = self.calc_bps(
                prev_tx_bps, now_tx_bytes, prev_tx_bytes, diff_sec)

            if (pg_current['tx_bps'].get(port) is not None
                    and pg_current['tx_pps'].get(port) is not None):
                pg_current['tx_bps_L1'][port] = calc_bps_L1(
                    pg_current['tx_bps'][port], pg_current['tx_pps'][port])
                pg_current['tx_bps_L1_lpf'][port] = calc_bps_L1(
                    pg_current['tx_bps_lpf'][port], pg_current['tx_pps_lpf'][port])
            else:
                pg_current['tx_bps_L1'][port] = None
                pg_current['tx_bps_L1_lpf'][port] = None

        # RX
        for port in list(pg_current['rx_pkts'].keys()):

            prev_rx_pps = pg_prev['rx_pps'].get(port)
            now_rx_pkts = pg_current['rx_pkts'].get(port)
            prev_rx_pkts = pg_prev['rx_pkts'].get(port)
            pg_current['rx_pps'][port], pg_current['rx_pps_lpf'][port] = self.calc_pps(
                prev_rx_pps, now_rx_pkts, prev_rx_pkts, diff_sec)

            prev_rx_bps = pg_prev['rx_bps'].get(port)
            now_rx_bytes = pg_current['rx_bytes'].get(port)
            prev_rx_bytes = pg_prev['rx_bytes'].get(port)
            pg_current['rx_bps'][port], pg_current['rx_bps_lpf'][port] = self.calc_bps(
                prev_rx_bps, now_rx_bytes, prev_rx_bytes, diff_sec)
            if (pg_current['rx_bps'].get(port) is not None
                    and pg_current['rx_pps'].get(port) is not None):
                pg_current['rx_bps_L1'][port] = calc_bps_L1(
                    pg_current['rx_bps'][port], pg_current['rx_pps'][port])
                pg_current['rx_bps_L1_lpf'][port] = calc_bps_L1(
                    pg_current['rx_bps_lpf'][port], pg_current['rx_pps_lpf'][port])
            else:
                pg_current['rx_bps_L1'][port] = None
                pg_current['rx_bps_L1_lpf'][port] = None

    def calc_pps(self, prev_bw, now, prev, diff_sec):
        return self.calc_bw(prev_bw, now, prev, diff_sec, False)

    def calc_bps(self, prev_bw, now, prev, diff_sec):
        return self.calc_bw(prev_bw, now, prev, diff_sec, True)

    # returns tuple - first value is real, second is low pass filtered
    def calc_bw(self, prev_bw, now, prev, diff_sec, is_bps):
        # B/W is not valid when the values are None
        if(now is None) or(prev is None):
            return(None, None)

        # calculate the B/W for current snapshot
        current_bw = (now - prev) / diff_sec
        if is_bps:
            current_bw *= 8

        # previous B/W is None ? ignore it
        if prev_bw is None:
            prev_bw = 0

        return(current_bw, 0.5 * prev_bw + 0.5 * current_bw)

    def _update(self, snapshot):
        # print(snapshot)
        # generate a new snapshot
        new_snapshot = self.process_snapshot(snapshot, self.latest_stats)

        # print new_snapshot
        # advance
        self.latest_stats = new_snapshot

        return True

    # for API

    def get_stats(self):
        stats = {}

        for pg_id, value in list(self.latest_stats.items()):
            # skip non ints
            if not is_intable(pg_id):
                # 'global' stats are in the same level of the pg_ids.
                # We do want them to go to the user
                if pg_id == 'global':
                    stats[pg_id] = value
                continue
            # bare counters
            stats[int(pg_id)] = {}
            for field in ['tx_pkts', 'tx_bytes', 'rx_pkts', 'rx_bytes']:
                val = self.get_rel([pg_id, field, 'total'])
                stats[int(pg_id)][field] = {
                    'total': val if val != 'N/A' else StatNotAvailable(field)}
                for port in list(value[field].keys()):
                    if is_intable(port):
                        val = self.get_rel([pg_id, field, port])
                        stats[int(pg_id)][field][int(
                            port)] = val if val != 'N/A' else StatNotAvailable(field)

            # BW values
            for field in ['tx_pps', 'tx_bps', 'tx_bps_L1', 'rx_pps', 'rx_bps', 'rx_bps_L1']:
                val = self.get([pg_id, field, 'total'])
                stats[int(pg_id)][field] = {
                    'total': val if val != 'N/A' else StatNotAvailable(field)}
                for port in list(value[field].keys()):
                    if is_intable(port):
                        val = self.get([pg_id, field, port])
                        stats[int(pg_id)][field][int(
                            port)] = val if val != 'N/A' else StatNotAvailable(field)

        return stats


class CUtilStats(CTRexStats):

    def __init__(self, client):
        super(CUtilStats, self).__init__()
        self.client = client
        self.history = deque(maxlen=1)
        self.mbuf_types_list = None
        self.last_update_ts = -999

    def get_stats(self, use_1sec_cache=False):
        time_now = time.time()
        if self.last_update_ts + 1 < time_now or not self.history or not use_1sec_cache:
            if self.client.is_connected():
                rc = self.client._transmit('get_utilization')
                if not rc:
                    raise STLError(rc)
                self.last_update_ts = time_now
                self.history.append(rc.data())
            else:
                self.history.append({})

        return self.history[-1]


class CXStats(CTRexStats):

    def __init__(self, client):
        super(CXStats, self).__init__()
        self.client = client
        self.names = []
        self.last_update_ts = -999

    def clear_stats(self, port_id=None):
        if port_id is None:
            ports = self.client.get_all_ports()
        elif type(port_id) is list:
            ports = port_id
        else:
            ports = [port_id]

        for port_id in ports:
            self.reference_stats[port_id] = self.get_stats(
                port_id, relative=False)

    def get_stats(self, port_id, use_1sec_cache=False, relative=True):
        time_now = time.time()
        if self.last_update_ts + 1 < time_now or not self.latest_stats or not use_1sec_cache:
            if self.client.is_connected():
                rc = self.client._transmit(
                    'get_port_xstats_values', params={'port_id': port_id})
                if not rc:
                    raise STLError(rc)
                self.last_update_ts = time_now
                values = rc.data().get('xstats_values', [])
                if len(values) != len(self.names):  # need to update names("keys")
                    rc = self.client._transmit(
                        'get_port_xstats_names', params={'port_id': port_id})
                    if not rc:
                        raise STLError(rc)
                    self.names = rc.data().get('xstats_names', [])
                if len(values) != len(self.names):
                    raise STLError('Length of get_xstats_names: %s '
                                   + 'and get_port_xstats_values: %s'
                                   % (len(self.names), len(values)))
                self.latest_stats[port_id] = OrderedDict(
                    [(key, val) for key, val in zip(self.names, values)])

        stats = OrderedDict()
        for key, val in list(self.latest_stats[port_id].items()):
            if relative:
                stats[key] = self.get_rel([port_id, key])
            else:
                stats[key] = self.get([port_id, key])
        return stats


class CPgIdStats(object):
    def __init__(self):
        self.reset()

    def reset(self):
        # sample when clear was last called. Values we return are last - ref
        self.ref = {'flow_stats': {}, 'latency': {}}
        # last sample values
        self.last = {'flow_stats': {}, 'latency': {}}

    def _get(self, src, field, default=None):
        if isinstance(field, list):
            # deep
            value = src
            for level in field:
                if level not in value:
                    return default
                value = value[level]
        else:
            # flat
            if field not in src:
                return default
            value = src[field]

        return value

    def get(self, field, format=False, suffix="", opts=None):
        value = self._get(self.last, field)
        if type(value) is StatNotAvailable:
            return 'N/A'
        if value is None:
            return 'N/A'

        return value if not format else format_num(value, suffix=suffix, opts=opts)

    def get_rel(self, field, format=False, suffix="", opts=None):
        value = self._get(self.last, field)
        if type(value) is StatNotAvailable:
            return 'N/A'
        if value is None:
            return 'N/A'

        base = self._get(self.ref, field, default=0)
        return(value - base) if not format else format_num(value - base, suffix=suffix, opts=opts)

    def clear_stats(self, clear_flow_stats=True, clear_latency_stats=True):
        keys_to_clear = []
        if clear_flow_stats:
            keys_to_clear.append('flow_stats')
        if clear_latency_stats:
            keys_to_clear.append('latency')

        for key in self.last:
            if key in keys_to_clear:
                self.ref[key].update(self.last[key])

        if 'latency' in self.ref and 'latency' in keys_to_clear:
            for pg_id in self.ref['latency']:
                if 'latency' in self.ref['latency'][pg_id]:
                    self.ref['latency'][pg_id]['latency']['total_max'] = 0

    def get_stats(self):
        flow_stat_fields = ['rx_pkts', 'tx_pkts', 'rx_bytes', 'tx_bytes']
        ret = copy.deepcopy(self.last)
        if 'flow_stats' not in ret:
            return ret
        for pg_id in ret['flow_stats']:
            for field in flow_stat_fields:
                if pg_id in self.ref['flow_stats'] and field in self.ref['flow_stats'][pg_id]:
                    for port in ret['flow_stats'][pg_id][field]:
                        if port in self.ref['flow_stats'][pg_id][field]:
                            # might be StatNotAvailable
                            try:
                                ret['flow_stats'][pg_id][field][port] -= (self.ref['flow_stats']
                                                                          [pg_id][field][port])
                            except Exception:
                                pass

        if 'latency' not in ret:
            return ret

        for pg_id in ret['latency']:
            if pg_id not in self.ref['latency']:
                continue
            if 'latency' in ret['latency'][pg_id]:
                to_delete = []
                for key in ret['latency'][pg_id]['latency']['histogram']:
                    if pg_id in self.ref['latency'] and key in (self.ref['latency'][pg_id]
                                                                ['latency']['histogram']):
                        ret['latency'][pg_id]['latency']['histogram'][key] -= (self.ref['latency']
                                                                               [pg_id]['latency']
                                                                               ['histogram'][key])
                        if ret['latency'][pg_id]['latency']['histogram'][key] == 0:
                            to_delete.append(key)
                # cleaning values in histogram which are 0(after decreasing ref from last)
                for del_key in to_delete:
                    del ret['latency'][pg_id]['latency']['histogram'][del_key]
                # special handling for 'total_max' field. We want to take it at first,
                # from server reported 'total_max'
                # After running clear_stats, we zero it, and start calculating ourselves
                # by looking at 'last_max' in each sampling
                if 'total_max' in self.ref['latency'][pg_id]['latency']:
                    ret['latency'][pg_id]['latency']['total_max'] = (self.ref['latency'][pg_id]
                                                                     ['latency']['total_max'])
                if 'err_cntrs' in ret['latency'][pg_id] and 'err_cntrs' in (self.ref['latency']
                                                                            [pg_id]):
                    for key in ret['latency'][pg_id]['err_cntrs']:
                        if key in self.ref['latency'][pg_id]['err_cntrs']:
                            ret['latency'][pg_id]['err_cntrs'][key] -= (self.ref['latency'][pg_id]
                                                                        ['err_cntrs'][key])
        return ret

    def save_stats(self, stats):
        # if pgid appears with different ver_id, delete it from the saved reference
        to_delete = []
        if 'ver_id' in self.ref:
            for pg_id in self.ref['ver_id']:
                if pg_id in stats['ver_id']:
                    if self.ref['ver_id'][pg_id] != stats['ver_id'][pg_id]:
                        to_delete.append(pg_id)
        else:
            self.ref['ver_id'] = {}

        for del_id in to_delete:
            try:
                del self.ref['flow_stats'][int(del_id)]
            except Exception:
                pass
            try:
                del self.ref['latency'][int(del_id)]
            except Exception:
                pass

        self.ref['ver_id'].update(stats['ver_id'])

        self.last = copy.deepcopy(stats)

        # update total_max
        if 'latency' in self.ref and 'latency' in self.last:
            for pg_id in self.ref['latency']:
                if 'latency' in self.ref['latency'][pg_id] and pg_id in self.last['latency']:
                    if 'total_max' in self.ref['latency'][pg_id]['latency']:
                        if (self.last['latency'][pg_id]
                            ['latency']['last_max']) > (self.ref['latency']
                                                        [pg_id]['latency']['total_max']):
                                (self.ref['latency'][pg_id]
                                 ['latency']['total_max']) = (self.last['latency'][pg_id]
                                                              ['latency']['last_max'])
                    else:
                        self.ref['latency'][pg_id]['latency']['total_max'] = self._get(
                            stats, ['latency', pg_id, 'latency', 'total_max'], default=0)
                    (self.last['latency'][pg_id]
                     ['latency']['total_max']) = self.ref['latency'][pg_id]['latency']['total_max']


if __name__ == "__main__":
    pass



#!/usr/bin/python3

from collections import OrderedDict

ON_OFF_DICT = OrderedDict([
    ('on', True),
    ('off', False),
])

UP_DOWN_DICT = OrderedDict([
    ('up', True),
    ('down', False),
])

FLOW_CTRL_DICT = OrderedDict([
    ('none', 0),     # Disable flow control
    ('tx',   1),     # Enable flowctrl on TX side(RX pause frames)
    ('rx',   2),     # Enable flowctrl on RX side(TX pause frames)
    ('full', 3),     # Enable flow control on both sides
])

# generate reverse dicts

for var_name in list(vars().keys()):
    if var_name.endswith('_DICT'):
        exec('{0}_REVERSED = OrderedDict([(val, key)for key, val in {0}.items()])'.format(var_name))



#!/usr/bin/python3

from .trex_stl_rx_service_api import RXServiceAPI

from trex_stl_lib.trex_stl_streams import STLStream, STLTXSingleBurst
from trex_stl_lib.trex_stl_packet_builder_scapy import *
from trex_stl_lib.trex_stl_types import *

from scapy.layers.l2 import Ether
from scapy.layers.inet6 import *
import time


class RXServiceIPv6(RXServiceAPI):

    def __init__(self, port, dst_ip, *a, **k):
        RXServiceAPI.__init__(self, port, *a, **k)
        self.attr = port.get_ts_attr()
        self.src_mac = self.attr['layer_cfg']['ether']['src']
        mac_for_ip = port.info.get('hw_mac', self.src_mac)
        self.src_ip = generate_ipv6(mac_for_ip)
        self.mld_ip = generate_ipv6_solicited_node(mac_for_ip)
        self.dst_ip = dst_ip
        self.responses = {}

    def pre_execute(self):
        return self.port.ok()

    def send_intermediate(self, streams):
        self.port.remove_all_streams()
        self.port.add_streams(streams)
        mult = {'op': 'abs', 'type': 'percentage', 'value': 100}
        self.port.start(mul=mult, force=True, duration=-1, mask=0xffffffff)

    def generate_ns(self, dst_mac, dst_ip):
        pkt = (Ether(src=self.src_mac, dst=dst_mac) /
               IPv6(src=self.src_ip, dst=dst_ip) /
               ICMPv6ND_NS(tgt=dst_ip) /
               ICMPv6NDOptSrcLLAddr(lladdr=self.src_mac))
        return STLStream(packet=STLPktBuilder(pkt=pkt), mode=STLTXSingleBurst(total_pkts=2))

    def generate_na(self, dst_mac, dst_ip):
        pkt = (Ether(src=self.src_mac, dst=dst_mac) /
               IPv6(src=self.src_ip, dst=dst_ip) /
               ICMPv6ND_NA(tgt=self.src_ip, R=0, S=1, O=1))
        return STLStream(packet=STLPktBuilder(pkt=pkt), mode=STLTXSingleBurst(total_pkts=2))

    def generate_ns_na(self, dst_mac, dst_ip):
        return [self.generate_ns(dst_mac, dst_ip), self.generate_na(dst_mac, dst_ip)]

    def execute(self, *a, **k):
        mult = self.attr['multicast']['enabled']
        try:
            if mult != True:
                # response might be multicast
                self.port.set_attr(multicast=True)
            return RXServiceAPI.execute(self, *a, **k)
        finally:
            if mult != True:
                self.port.set_attr(multicast=False)


class RXServiceIPv6Scan(RXServiceIPv6):
    ''' Ping with given IPv6(usually all nodes address) and wait for responses until timeout '''

    def get_name(self):
        return 'IPv6 scanning'

    def generate_request(self):
        dst_mac = multicast_mac_from_ipv6(self.dst_ip)
        dst_mld_ip = 'ff02::16'
        dst_mld_mac = multicast_mac_from_ipv6(dst_mld_ip)

        mld_pkt = (Ether(src=self.src_mac, dst=dst_mld_mac) /
                   IPv6(src=self.src_ip, dst=dst_mld_ip, hlim=1) /
                   IPv6ExtHdrHopByHop(options=[RouterAlert(), PadN()]) /
                   ICMPv6MLReportV2() /
                   MLDv2Addr(type=4, len=0, multicast_addr='ff02::2'))
        ping_pkt = (Ether(src=self.src_mac, dst=dst_mac) /
                    IPv6(src=self.src_ip, dst=self.dst_ip, hlim=1) /
                    ICMPv6EchoRequest())

        mld_stream = STLStream(packet=STLPktBuilder(pkt=mld_pkt),
                               mode=STLTXSingleBurst(total_pkts=2))
        ping_stream = STLStream(packet=STLPktBuilder(pkt=ping_pkt),
                                mode=STLTXSingleBurst(total_pkts=2))
        return [mld_stream, ping_stream]

    def on_pkt_rx(self, pkt, start_ts):
        # convert to scapy
        scapy_pkt = Ether(pkt['binary'])

        if scapy_pkt.haslayer('ICMPv6ND_NS') and scapy_pkt.haslayer('ICMPv6NDOptSrcLLAddr'):
            node_mac = scapy_pkt.getlayer(ICMPv6NDOptSrcLLAddr).lladdr
            node_ip = scapy_pkt.getlayer(IPv6).src
            if node_ip not in self.responses:
                self.send_intermediate(self.generate_ns_na(node_mac, node_ip))

        elif scapy_pkt.haslayer('ICMPv6ND_NA'):
            is_router = scapy_pkt.getlayer(ICMPv6ND_NA).R
            node_ip = scapy_pkt.getlayer(ICMPv6ND_NA).tgt
            dst_ip = scapy_pkt.getlayer(IPv6).dst
            node_mac = scapy_pkt.src
            if node_ip not in self.responses and dst_ip == self.src_ip:
                self.responses[node_ip] = {
                    'type': 'Router' if is_router else 'Host', 'mac': node_mac}

        elif scapy_pkt.haslayer('ICMPv6EchoReply'):
            node_mac = scapy_pkt.src
            node_ip = scapy_pkt.getlayer(IPv6).src
            if node_ip == self.dst_ip:
                return self.port.ok([{'ipv6': node_ip, 'mac': node_mac}])
            if node_ip not in self.responses:
                self.send_intermediate(self.generate_ns_na(node_mac, node_ip))

    def on_timeout(self):
        return self.port.ok([{'type': v['type'], 'mac':  v['mac'], 'ipv6': k} for k, v in list(self.responses.items())])


class RXServiceICMPv6(RXServiceIPv6):
    '''
    Ping some IPv6 location.
    If the dest MAC is found from scanning, use it.
    Otherwise, send to default port dest.
    '''

    def __init__(self, port, pkt_size, dst_mac=None, *a, **k):
        super(RXServiceICMPv6, self).__init__(port, *a, **k)
        self.pkt_size = pkt_size
        self.dst_mac = dst_mac
        self.result = {}

    def get_name(self):
        return 'PING6'

    def pre_execute(self):
        if self.dst_mac is None and not self.port.is_resolved():
            return self.port.err('ping - port has an unresolved destination, cannot determine next hop MAC address')
        return self.port.ok()

    # return a list of streams for request

    def generate_request(self):
        attrs = self.port.get_ts_attr()

        if self.dst_mac is None:
            self.dst_mac = attrs['layer_cfg']['ether']['dst']

        ping_pkt = (Ether(src=self.src_mac, dst=self.dst_mac) /
                    IPv6(src=self.src_ip, dst=self.dst_ip) /
                    ICMPv6EchoRequest())
        pad = max(0, self.pkt_size - len(ping_pkt))
        ping_pkt /= pad * 'x'

        return STLStream(packet=STLPktBuilder(pkt=ping_pkt), mode=STLTXSingleBurst(total_pkts=2))

    def on_pkt_rx(self, pkt, start_ts):
        scapy_pkt = Ether(pkt['binary'])

        if scapy_pkt.haslayer('ICMPv6EchoReply'):
            node_ip = scapy_pkt.getlayer(IPv6).src
            hlim = scapy_pkt.getlayer(IPv6).hlim
            dst_ip = scapy_pkt.getlayer(IPv6).dst
            if dst_ip != self.src_ip:  # not our ping
                return

            dt = pkt['ts'] - start_ts
            self.result['formatted_string'] = 'Reply from {0}: bytes={1}, time={2:.2f}ms, hlim={3}'.format(
                node_ip, len(pkt['binary']), dt * 1000, hlim)
            self.result['src_ip'] = node_ip
            self.result['rtt'] = dt * 1000
            self.result['ttl'] = hlim
            self.result['status'] = 'success'
            return self.port.ok(self.result)

        if scapy_pkt.haslayer('ICMPv6ND_NS') and scapy_pkt.haslayer('ICMPv6NDOptSrcLLAddr'):
            node_mac = scapy_pkt.getlayer(ICMPv6NDOptSrcLLAddr).lladdr
            node_ip = scapy_pkt.getlayer(IPv6).src
            dst_ip = scapy_pkt.getlayer(IPv6).dst
            if dst_ip != self.src_ip:  # not our ping
                return
            self.send_intermediate(self.generate_ns_na(node_mac, node_ip))

        if scapy_pkt.haslayer('ICMPv6DestUnreach'):
            node_ip = scapy_pkt.getlayer(IPv6).src
            dst_ip = scapy_pkt.getlayer(IPv6).dst
            if dst_ip != self.src_ip:  # not our ping
                return
            self.result['formatted_string'] = 'Reply from {0}: Destination host unreachable'.format(
                node_ip)
            self.result['status'] = 'unreachable'
            return self.port.ok(self.result)

    # return the str of a timeout err
    def on_timeout(self):
        self.result['formatted_string'] = 'Request timed out.'
        self.result['status'] = 'timeout'
        return self.port.ok(self.result)
